"""V2.0 present-proof anoncreds presentation-exchange format handler."""

import json
import logging
from typing import Mapping, Optional, Tuple

from marshmallow import RAISE

from ......anoncreds.holder import AnonCredsHolder
from ......anoncreds.models.predicate import Predicate
from ......anoncreds.models.presentation_request import AnonCredsPresentationRequestSchema
from ......anoncreds.models.proof import AnonCredsProofSchema
from ......anoncreds.models.utils import get_requested_creds_from_proof_request_preview
from ......anoncreds.registry import AnonCredsRegistry
from ......anoncreds.util import generate_pr_nonce
from ......anoncreds.verifier import AnonCredsVerifier
from ......messaging.decorators.attach_decorator import AttachDecorator
from ......messaging.util import canon
from ....anoncreds.pres_exch_handler import AnonCredsPresExchHandler
from ...message_types import ATTACHMENT_FORMAT, PRES_20, PRES_20_PROPOSAL, PRES_20_REQUEST
from ...messages.pres import V20Pres
from ...messages.pres_format import V20PresFormat
from ...models.pres_exchange import V20PresExRecord
from ..handler import V20PresFormatHandler, V20PresFormatHandlerError

LOGGER = logging.getLogger(__name__)


class AnonCredsPresExchangeHandler(V20PresFormatHandler):
    """AnonCreds presentation format handler."""

    format = V20PresFormat.Format.ANONCREDS

    @classmethod
    def validate_fields(cls, message_type: str, attachment_data: Mapping):
        """Validate attachment data for a specific message type.

        Uses marshmallow schemas to validate if format specific attachment data
        is valid for the specified message type. Only does structural and type
        checks, does not validate if .e.g. the issuer value is valid.


        Args:
            message_type (str): The message type to validate the attachment data for.
                Should be one of the message types as defined in message_types.py
            attachment_data (Mapping): [description]
                The attachment data to validate

        Raises:
            Exception: When the data is not valid.

        """
        mapping = {
            PRES_20_REQUEST: AnonCredsPresentationRequestSchema,
            PRES_20_PROPOSAL: AnonCredsPresentationRequestSchema,
            PRES_20: AnonCredsProofSchema,
        }

        # Get schema class
        Schema = mapping[message_type]

        # Validate, throw if not valid
        Schema(unknown=RAISE).load(attachment_data)

    def get_format_identifier(self, message_type: str) -> str:
        """Get attachment format identifier for format and message combination.

        Args:
            message_type (str): Message type for which to return the format identifier

        Returns:
            str: Issue credential attachment format identifier

        """
        return ATTACHMENT_FORMAT[message_type][AnonCredsPresExchangeHandler.format.api]

    def get_format_data(
        self, message_type: str, data: dict
    ) -> Tuple[V20PresFormat, AttachDecorator]:
        """Get presentation format and attach objects for use in pres_ex messages."""
        return (
            V20PresFormat(
                attach_id=AnonCredsPresExchangeHandler.format.api,
                format_=self.get_format_identifier(message_type),
            ),
            AttachDecorator.data_base64(
                data,
                ident=AnonCredsPresExchangeHandler.format.api,
            ),
        )

    async def create_bound_request(
        self,
        pres_ex_record: V20PresExRecord,
        request_data: Optional[dict] = None,
    ) -> Tuple[V20PresFormat, AttachDecorator]:
        """Create a presentation request bound to a proposal.

        Args:
            pres_ex_record: Presentation exchange record for which
                to create presentation request
            request_data: Dict

        Returns:
            A tuple (updated presentation exchange record, presentation request message)

        """
        proof_request = pres_ex_record.pres_proposal.attachment(
            AnonCredsPresExchangeHandler.format
        )
        if request_data:
            proof_request["name"] = request_data.get("name", "proof-request")
            proof_request["version"] = request_data.get("version", "1.0")
            proof_request["nonce"] = (
                request_data.get("nonce") or await generate_pr_nonce()
            )
        else:
            proof_request["name"] = "proof-request"
            proof_request["version"] = "1.0"
            proof_request["nonce"] = await generate_pr_nonce()
        return self.get_format_data(PRES_20_REQUEST, proof_request)

    async def create_pres(
        self,
        pres_ex_record: V20PresExRecord,
        request_data: Optional[dict] = None,
    ) -> Tuple[V20PresFormat, AttachDecorator]:
        """Create a presentation."""
        requested_credentials = {}

        # This is used for the fallback to indy format
        from ..indy.handler import IndyPresExchangeHandler

        if not request_data:
            try:
                proof_request = pres_ex_record.pres_request
                # Fall back to indy format should be removed when indy format retired
                proof_request = proof_request.attachment(
                    AnonCredsPresExchangeHandler.format
                ) or proof_request.attachment(IndyPresExchangeHandler.format)
                requested_credentials = (
                    await get_requested_creds_from_proof_request_preview(
                        proof_request,
                        holder=AnonCredsHolder(self._profile),
                    )
                )
            except ValueError as err:
                LOGGER.warning(f"{err}")
                raise V20PresFormatHandlerError(f"No matching credentials found: {err}")
        else:
            # Fall back to indy format should be removed when indy format retired
            if (
                AnonCredsPresExchangeHandler.format.api in request_data
                or IndyPresExchangeHandler.format.api in request_data
            ):
                spec = request_data.get(
                    AnonCredsPresExchangeHandler.format.api
                ) or request_data.get(IndyPresExchangeHandler.format.api)
                requested_credentials = {
                    "self_attested_attributes": spec["self_attested_attributes"],
                    "requested_attributes": spec["requested_attributes"],
                    "requested_predicates": spec["requested_predicates"],
                }
        handler = AnonCredsPresExchHandler(self._profile)
        presentation_proof = await handler.return_presentation(
            pres_ex_record=pres_ex_record,
            requested_credentials=requested_credentials,
        )

        # This is used for the fallback to indy format. Should be removed when indy
        # format retired
        if request_data.get("indy"):
            return IndyPresExchangeHandler(self.profile).get_format_data(
                PRES_20, presentation_proof
            )
        return self.get_format_data(PRES_20, presentation_proof)

    async def receive_pres(self, message: V20Pres, pres_ex_record: V20PresExRecord):
        """Receive a presentation and check for presented values vs. proposal request."""

        async def _check_proof_vs_proposal():
            """Check for bait and switch in presented values vs. proposal request."""
            from ..indy.handler import IndyPresExchangeHandler

            # Fall back to indy format should be removed when indy format retired
            proof_req = pres_ex_record.pres_request.attachment(
                AnonCredsPresExchangeHandler.format
            ) or pres_ex_record.pres_request.attachment(IndyPresExchangeHandler.format)

            # revealed attrs
            for reft, attr_spec in proof["requested_proof"]["revealed_attrs"].items():
                proof_req_attr_spec = proof_req["requested_attributes"].get(reft)
                if not proof_req_attr_spec:
                    raise V20PresFormatHandlerError(
                        f"Presentation referent {reft} not in proposal request"
                    )
                req_restrictions = proof_req_attr_spec.get("restrictions", {})

                name = proof_req_attr_spec["name"]
                proof_value = attr_spec["raw"]
                sub_proof_index = attr_spec["sub_proof_index"]
                schema_id = proof["identifiers"][sub_proof_index]["schema_id"]
                cred_def_id = proof["identifiers"][sub_proof_index]["cred_def_id"]
                registry = self.profile.inject(AnonCredsRegistry)
                schema = await registry.get_schema(self.profile, schema_id)
                cred_def = await registry.get_credential_definition(
                    self.profile, cred_def_id
                )
                criteria = {
                    "schema_id": schema_id,
                    "schema_issuer_did": schema.schema_value.issuer_id,
                    "schema_name": schema.schema_value.name,
                    "schema_version": schema.schema_value.version,
                    "cred_def_id": cred_def_id,
                    "issuer_did": cred_def.credential_definition.issuer_id,
                    f"attr::{name}::value": proof_value,
                }

                if (
                    not any(r.items() <= criteria.items() for r in req_restrictions)
                    and len(req_restrictions) != 0
                ):
                    raise V20PresFormatHandlerError(
                        f"Presented attribute {reft} does not satisfy proof request "
                        f"restrictions {req_restrictions}"
                    )

            # revealed attr groups
            for reft, attr_spec in (
                proof["requested_proof"].get("revealed_attr_groups", {}).items()
            ):
                proof_req_attr_spec = proof_req["requested_attributes"].get(reft)
                if not proof_req_attr_spec:
                    raise V20PresFormatHandlerError(
                        f"Presentation referent {reft} not in proposal request"
                    )
                req_restrictions = proof_req_attr_spec.get("restrictions", {})
                proof_values = {
                    name: values["raw"] for name, values in attr_spec["values"].items()
                }
                sub_proof_index = attr_spec["sub_proof_index"]
                schema_id = proof["identifiers"][sub_proof_index]["schema_id"]
                cred_def_id = proof["identifiers"][sub_proof_index]["cred_def_id"]
                registry = self.profile.inject(AnonCredsRegistry)
                schema = await registry.get_schema(self.profile, schema_id)
                cred_def = await registry.get_credential_definition(
                    self.profile, cred_def_id
                )
                criteria = {
                    "schema_id": schema_id,
                    "schema_issuer_did": schema.schema_value.issuer_id,
                    "schema_name": schema.schema_value.name,
                    "schema_version": schema.schema_value.version,
                    "cred_def_id": cred_def_id,
                    "issuer_did": cred_def.credential_definition.issuer_id,
                    **{
                        f"attr::{name}::value": value
                        for name, value in proof_values.items()
                    },
                }

                if (
                    not any(r.items() <= criteria.items() for r in req_restrictions)
                    and len(req_restrictions) != 0
                ):
                    raise V20PresFormatHandlerError(
                        f"Presented attr group {reft} does not satisfy proof request "
                        f"restrictions {req_restrictions}"
                    )

            # predicate bounds
            for reft, pred_spec in proof["requested_proof"]["predicates"].items():
                proof_req_pred_spec = proof_req["requested_predicates"].get(reft)
                if not proof_req_pred_spec:
                    raise V20PresFormatHandlerError(
                        f"Presentation referent {reft} not in proposal request"
                    )
                req_name = proof_req_pred_spec["name"]
                req_pred = Predicate.get(proof_req_pred_spec["p_type"])
                req_value = proof_req_pred_spec["p_value"]
                req_restrictions = proof_req_pred_spec.get("restrictions", {})
                for req_restriction in req_restrictions:
                    for k in list(req_restriction):  # cannot modify en passant
                        if k.startswith("attr::"):
                            # let anoncreds-sdk reject mismatch here
                            req_restriction.pop(k)
                sub_proof_index = pred_spec["sub_proof_index"]
                for ge_proof in proof["proof"]["proofs"][sub_proof_index][
                    "primary_proof"
                ]["ge_proofs"]:
                    proof_pred_spec = ge_proof["predicate"]
                    if proof_pred_spec["attr_name"] != canon(req_name):
                        continue
                    if not (
                        Predicate.get(proof_pred_spec["p_type"]) is req_pred
                        and proof_pred_spec["value"] == req_value
                    ):
                        raise V20PresFormatHandlerError(
                            f"Presentation predicate on {req_name} "
                            "mismatches proposal request"
                        )
                    break
                else:
                    raise V20PresFormatHandlerError(
                        f"Proposed request predicate on {req_name} not in presentation"
                    )

                schema_id = proof["identifiers"][sub_proof_index]["schema_id"]
                cred_def_id = proof["identifiers"][sub_proof_index]["cred_def_id"]
                registry = self.profile.inject(AnonCredsRegistry)
                schema = await registry.get_schema(self.profile, schema_id)
                cred_def = await registry.get_credential_definition(
                    self.profile, cred_def_id
                )
                criteria = {
                    "schema_id": schema_id,
                    "schema_issuer_did": schema.schema_value.issuer_id,
                    "schema_name": schema.schema_value.name,
                    "schema_version": schema.schema_value.version,
                    "cred_def_id": cred_def_id,
                    "issuer_did": cred_def.credential_definition.issuer_id,
                }

                if (
                    not any(r.items() <= criteria.items() for r in req_restrictions)
                    and len(req_restrictions) != 0
                ):
                    raise V20PresFormatHandlerError(
                        f"Presented predicate {reft} does not satisfy proof request "
                        f"restrictions {req_restrictions}"
                    )

        proof = message.attachment(AnonCredsPresExchangeHandler.format)
        await _check_proof_vs_proposal()

    async def verify_pres(self, pres_ex_record: V20PresExRecord) -> V20PresExRecord:
        """Verify a presentation.

        Args:
            pres_ex_record: presentation exchange record
                with presentation request and presentation to verify

        Returns:
            presentation exchange record, updated

        """
        pres_request_msg = pres_ex_record.pres_request
        proof_request = pres_request_msg.attachment(AnonCredsPresExchangeHandler.format)
        proof = pres_ex_record.pres.attachment(AnonCredsPresExchangeHandler.format)
        verifier = AnonCredsVerifier(self._profile)

        (
            schemas,
            cred_defs,
            rev_reg_defs,
            rev_lists,
        ) = await verifier.process_pres_identifiers(proof["identifiers"])

        verifier = AnonCredsVerifier(self._profile)

        (verified, verified_msgs) = await verifier.verify_presentation(
            proof_request,
            proof,
            schemas,
            cred_defs,
            rev_reg_defs,
            rev_lists,
        )
        pres_ex_record.verified = json.dumps(verified)
        pres_ex_record.verified_msgs = list(set(verified_msgs))
        return pres_ex_record



from unittest import TestCase

from ..cred import IndyAttrValue, IndyCredential
from ..cred_abstract import IndyCredAbstract, IndyKeyCorrectnessProof
from ..cred_request import IndyCredRequest

KC_PROOF = {
    "c": "123467890",
    "xz_cap": "12345678901234567890",
    "xr_cap": [
        [
            "remainder",
            "1234567890",
        ],
        [
            "number",
            "12345678901234",
        ],
        [
            "master_secret",
            "12345678901234",
        ],
    ],
}

TEST_DID = "LjgpST2rjsoxYegQDRm7EL"
SCHEMA_NAME = "bc-reg"
SCHEMA_TXN = 12
SCHEMA_ID = f"{TEST_DID}:2:{SCHEMA_NAME}:1.0"
CRED_DEF_ID = f"{TEST_DID}:3:CL:12:default"


class TestIndyKeyCorrectnessProof(TestCase):
    """Indy key correctness proof tests"""

    def test_serde(self):
        """Test de/serialization."""
        kcp = IndyKeyCorrectnessProof.deserialize(KC_PROOF)
        assert type(kcp) is IndyKeyCorrectnessProof

        kcp_dict = kcp.serialize()
        assert kcp_dict == KC_PROOF


class TestIndyCredAbstract(TestCase):
    """Indy cred abstract tests."""

    def test_serde(self):
        """Test de/serialization."""
        obj = {
            "schema_id": SCHEMA_ID,
            "cred_def_id": CRED_DEF_ID,
            "nonce": "1234567890",
            "key_correctness_proof": KC_PROOF,
        }
        cred_abstract = IndyCredAbstract.deserialize(obj)
        assert type(cred_abstract) is IndyCredAbstract

        cred_abstract_dict = cred_abstract.serialize()
        assert cred_abstract_dict == obj


class TestIndyCredRequest(TestCase):
    """Indy cred request tests."""

    def test_serde(self):
        """Test de/serialization."""
        obj = {
            "prover_did": f"did:sov:{TEST_DID}",
            "cred_def_id": CRED_DEF_ID,
            "blinded_ms": {"...": "..."},
            "blinded_ms_correctness_proof": {"...": "..."},
            "nonce": "1234567890",
        }
        cred_request = IndyCredRequest.deserialize(obj)
        assert type(cred_request) is IndyCredRequest

        cred_request_dict = cred_request.serialize()
        assert cred_request_dict == obj


class TestIndyAttrValue(TestCase):
    """Indy attr value tests."""

    def test_serde(self):
        """Test de/serialization."""
        obj = {
            "raw": "test",
            "encoded": "1234567890",
        }
        attr_val = IndyAttrValue.deserialize(obj)
        assert type(attr_val) is IndyAttrValue

        attr_val_dict = attr_val.serialize()
        assert attr_val_dict == obj


class TestIndyCredential(TestCase):
    """Indy credential tests."""

    def test_serde(self):
        """Test de/serialization."""
        obj = {
            "schema_id": SCHEMA_ID,
            "cred_def_id": CRED_DEF_ID,
            "rev_reg_id": None,
            "values": {
                "busId": {
                    "raw": "12345",
                    "encoded": "12345",
                },
                "legalName": {
                    "raw": "Muffin Moon",
                    "encoded": "13419834198651328645901659586128164",
                },
            },
            "signature": {"...": "..."},
            "signature_correctness_proof": {"...": "..."},
            "rev_reg": None,
            "witness": None,
        }
        cred = IndyCredential.deserialize(obj)
        assert type(cred) is IndyCredential

        cred_dict = cred.serialize()
        assert cred_dict.items() <= obj.items()



"""The Delivery Queue.

The delivery queue holds and manages messages that have not yet
been delivered to their intended destination.

"""

import time

from ..outbound.message import OutboundMessage


class QueuedMessage:
    """Wrapper Class for queued messages.

    Allows tracking Metadata.
    """

    def __init__(self, msg: OutboundMessage):
        """Create Wrapper for queued message.

        Automatically sets timestamp on create.
        """
        self.msg = msg
        self.timestamp = time.time()

    def older_than(self, compare_timestamp: float) -> bool:
        """Age Comparison.

        Allows you to test age as compared to the provided timestamp.

        Args:
            compare_timestamp: The timestamp to compare
        """
        return self.timestamp < compare_timestamp


class DeliveryQueue:
    """DeliveryQueue class.

    Manages undelivered messages.
    """

    def __init__(self) -> None:
        """Initialize an instance of DeliveryQueue.

        This uses an in memory structure to queue messages.
        """

        self.queue_by_key = {}
        self.ttl_seconds = 604800  # one week

    def expire_messages(self, ttl=None):
        """Expire messages that are past the time limit.

        Args:
            ttl: Optional. Allows override of configured ttl
        """

        ttl_seconds = ttl or self.ttl_seconds
        horizon = time.time() - ttl_seconds
        for key in self.queue_by_key.keys():
            self.queue_by_key[key] = [
                wm for wm in self.queue_by_key[key] if not wm.older_than(horizon)
            ]

    def add_message(self, msg: OutboundMessage):
        """Add an OutboundMessage to delivery queue.

        The message is added once per recipient key

        Args:
            msg: The OutboundMessage to add
        """
        keys = set()
        if msg.target:
            keys.update(msg.target.recipient_keys)
        if msg.reply_to_verkey:
            keys.add(msg.reply_to_verkey)
        wrapped_msg = QueuedMessage(msg)
        for recipient_key in keys:
            if recipient_key not in self.queue_by_key:
                self.queue_by_key[recipient_key] = []
            self.queue_by_key[recipient_key].append(wrapped_msg)

    def has_message_for_key(self, key: str):
        """Check for queued messages by key.

        Args:
            key: The key to use for lookup
        """
        if key in self.queue_by_key and len(self.queue_by_key[key]):
            return True
        return False

    def message_count_for_key(self, key: str):
        """Count of queued messages by key.

        Args:
            key: The key to use for lookup
        """
        if key in self.queue_by_key:
            return len(self.queue_by_key[key])
        else:
            return 0

    def get_one_message_for_key(self, key: str):
        """Remove and return a matching message.

        Args:
            key: The key to use for lookup
        """
        if key in self.queue_by_key:
            return self.queue_by_key[key].pop(0).msg

    def inspect_all_messages_for_key(self, key: str):
        """Return all messages for key.

        Args:
            key: The key to use for lookup
        """
        if key in self.queue_by_key:
            for wrapped_msg in self.queue_by_key[key]:
                yield wrapped_msg.msg

    def remove_message_for_key(self, key: str, msg: OutboundMessage):
        """Remove specified message from queue for key.

        Args:
            key: The key to use for lookup
            msg: The message to remove from the queue
        """
        if key in self.queue_by_key:
            for wrapped_msg in self.queue_by_key[key]:
                if wrapped_msg.msg == msg:
                    self.queue_by_key[key].remove(wrapped_msg)
                    if not self.queue_by_key[key]:
                        del self.queue_by_key[key]
                    break  # exit processing loop



import http
import os
from unittest import IsolatedAsyncioTestCase

import pytest
from anoncreds import (
    AnoncredsError,
    AnoncredsErrorCode,
    Credential,
    CredentialDefinition,
    RevocationRegistryDefinition,
    RevocationRegistryDefinitionPrivate,
    RevocationStatusList,
    Schema,
)
from aries_askar import AskarError, AskarErrorCode
from requests import RequestException, Session

from ...anoncreds.issuer import AnonCredsIssuer
from ...anoncreds.models.credential_definition import CredDef
from ...anoncreds.models.revocation import (
    RevList,
    RevListResult,
    RevListState,
    RevRegDef,
    RevRegDefResult,
    RevRegDefState,
    RevRegDefValue,
)
from ...anoncreds.models.schema import AnonCredsSchema, GetSchemaResult
from ...anoncreds.registry import AnonCredsRegistry
from ...anoncreds.tests.mock_objects import MOCK_REV_REG_DEF
from ...anoncreds.tests.test_issuer import MockCredDefEntry
from ...askar.profile_anon import AskarAnonCredsProfileSession
from ...core.event_bus import Event, EventBus, MockEventBus
from ...tails.anoncreds_tails_server import AnonCredsTailsServer
from ...tests import mock
from ...utils.testing import create_test_profile
from .. import revocation as test_module

rev_reg_def = RevRegDef(
    tag="tag",
    cred_def_id="CsQY9MGeD3CQP4EyuVFo5m:3:CL:14951:MYCO_Biomarker",
    value=RevRegDefValue(
        max_cred_num=100,
        public_keys={
            "accum_key": {"z": "1 0BB...386"},
        },
        tails_hash="58NNWYnVxVFzAfUztwGSNBL4551XNq6nXk56pCiKJxxt",
        tails_location="https://tails-server.com",
    ),
    issuer_id="CsQY9MGeD3CQP4EyuVFo5m",
    type="CL_ACCUM",
)

rev_list = RevList(
    issuer_id="CsQY9MGeD3CQP4EyuVFo5m",
    current_accumulator="21 124C594B6B20E41B681E92B2C43FD165EA9E68BC3C9D63A82C8893124983CAE94 21 124C5341937827427B0A3A32113BD5E64FB7AB39BD3E5ABDD7970874501CA4897 6 5438CB6F442E2F807812FD9DC0C39AFF4A86B1E6766DBB5359E86A4D70401B0F 4 39D1CA5C4716FFC4FE0853C4FF7F081DFD8DF8D2C2CA79705211680AC77BF3A1 6 70504A5493F89C97C225B68310811A41AD9CD889301F238E93C95AD085E84191 4 39582252194D756D5D86D0EED02BF1B95CE12AED2FA5CD3C53260747D891993C",
    revocation_list=[0, 1, 1, 0],
    timestamp=1669640864487,
    rev_reg_def_id="4xE68b6S5VRFrKMMG1U95M:4:4xE68b6S5VRFrKMMG1U95M:3:CL:59232:default:CL_ACCUM:4ae1cc6c-f6bd-486c-8057-88f2ce74e960",
)


class MockRevRegDefEntry:
    def __init__(self, name="name"):
        self.name = name

    rev_reg_def_id = "test-rev-reg-def-id"

    tags = {
        "state": RevRegDefState.STATE_ACTION,
    }
    value = "mock_value"
    value_json = {
        "value": {
            "maxCredNum": 100,
            "publicKeys": {"accumKey": {"z": "1 0BB...386"}},
            "tailsHash": "string",
            "tailsLocation": "string",
        },
        "credDefId": "CsQY9MGeD3CQP4EyuVFo5m:3:CL:14951:MYCO_Biomarker",
        "issuerId": "CsQY9MGeD3CQP4EyuVFo5m",
        "revocDefType": "CL_ACCUM",
        "tag": "string",
    }


class MockEntry:
    def __init__(
        self, name="name", value_json="", raw_value="raw-value", value="value", tags={}
    ) -> None:
        self.name = name
        self.value_json = value_json
        self.raw_value = raw_value
        self.value = value
        self.tags = tags


class MockRevListEntry:
    tags = {
        "state": RevListState.STATE_ACTION,
    }
    value = "mock_value"
    value_json = {
        "rev_list": {
            "issuerId": "CsQY9MGeD3CQP4EyuVFo5m",
            "revRegDefId": "4xE68b6S5VRFrKMMG1U95M:4:4xE68b6S5VRFrKMMG1U95M:3:CL:59232:default:CL_ACCUM:4ae1cc6c-f6bd-486c-8057-88f2ce74e960",
            "revocationList": [0, 1, 1, 0],
            "currentAccumulator": "21 124C594B6B20E41B681E92B2C43FD165EA9E68BC3C9D63A82C8893124983CAE94 21 124C5341937827427B0A3A32113BD5E64FB7AB39BD3E5ABDD7970874501CA4897 6 5438CB6F442E2F807812FD9DC0C39AFF4A86B1E6766DBB5359E86A4D70401B0F 4 39D1CA5C4716FFC4FE0853C4FF7F081DFD8DF8D2C2CA79705211680AC77BF3A1 6 70504A5493F89C97C225B68310811A41AD9CD889301F238E93C95AD085E84191 4 39582252194D756D5D86D0EED02BF1B95CE12AED2FA5CD3C53260747D891993C",
            "timestamp": 1669640864487,
        }
    }


@pytest.mark.anoncreds
class TestAnonCredsRevocation(IsolatedAsyncioTestCase):
    async def asyncSetUp(self) -> None:
        self.profile = await create_test_profile(
            settings={
                "wallet.type": "askar-anoncreds",
                "tails_server_base_url": "http://tails-server.com",
            },
        )
        self.revocation = test_module.AnonCredsRevocation(self.profile)

    async def test_init(self):
        assert self.revocation.profile == self.profile

    async def test_notify(self):
        self.profile.inject = mock.Mock(return_value=MockEventBus())
        await self.revocation.notify(Event(topic="test-topic"))

    async def test_create_and_register_revocation_registry_definition_fails_to_get_cred_def(
        self,
    ):
        # AnonCreds error
        with self.assertRaises(test_module.AnonCredsRevocationError):
            await self.revocation.create_and_register_revocation_registry_definition(
                issuer_id="test-issuer-id",
                cred_def_id="test-cred-def-id",
                registry_type="test-registry-type",
                tag="test-tag",
                max_cred_num=100,
            )
        # fetch returns None
        with self.assertRaises(test_module.AnonCredsRevocationError):
            await self.revocation.create_and_register_revocation_registry_definition(
                issuer_id="test-issuer-id",
                cred_def_id="test-cred-def-id",
                registry_type="test-registry-type",
                tag="test-tag",
                max_cred_num=100,
            )

    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    @mock.patch.object(
        test_module.AnonCredsRevocation,
        "generate_public_tails_uri",
        return_value="https://tails.uri",
    )
    @mock.patch.object(
        test_module.AnonCredsRevocation,
        "notify",
        return_value=None,
    )
    async def test_create_and_register_revocation_registry_definition(
        self, mock_notify, mock_tails_uri, mock_handle
    ):
        self.profile.inject = mock.Mock(
            return_value=mock.MagicMock(
                register_revocation_registry_definition=mock.CoroutineMock(
                    return_value=RevRegDefResult(
                        job_id="test-job-id",
                        revocation_registry_definition_state=RevRegDefState(
                            state=RevRegDefState.STATE_FINISHED,
                            revocation_registry_definition_id="active-reg-reg",
                            revocation_registry_definition=RevRegDef(
                                tag="tag",
                                cred_def_id="CsQY9MGeD3CQP4EyuVFo5m:3:CL:14951:MYCO_Biomarker",
                                value=RevRegDefValue(
                                    max_cred_num=100,
                                    public_keys={
                                        "accum_key": {"z": "1 0BB...386"},
                                    },
                                    tails_hash="58NNWYnVxVFzAfUztwGSNBL4551XNq6nXk56pCiKJxxt",
                                    tails_location="http://tails-server.com",
                                ),
                                issuer_id="CsQY9MGeD3CQP4EyuVFo5m",
                                type="CL_ACCUM",
                            ),
                        ),
                        registration_metadata={},
                        revocation_registry_definition_metadata={},
                    )
                )
            )
        )
        self.profile.transaction = mock.Mock(
            return_value=mock.MagicMock(
                insert=mock.CoroutineMock(return_value=None),
                commit=mock.CoroutineMock(return_value=None),
            )
        )
        schema = Schema.create(
            name="MYCO Biomarker",
            attr_names=["biomarker_id"],
            issuer_id="did:indy:sovrin:SGrjRL82Y9ZZbzhUDXokvQ",
            version="1.0",
        )

        (cred_def, _, _) = CredentialDefinition.create(
            schema_id="CsQY9MGeD3CQP4EyuVFo5m:2:MYCO Biomarker:0.0.3",
            schema=schema,
            issuer_id="did:indy:sovrin:SGrjRL82Y9ZZbzhUDXokvQ",
            tag="tag",
            support_revocation=True,
            signature_type="CL",
        )
        mock_handle.fetch = mock.CoroutineMock(
            return_value=MockEntry(raw_value=cred_def.to_json_buffer())
        )

        result = await self.revocation.create_and_register_revocation_registry_definition(
            issuer_id="did:indy:sovrin:SGrjRL82Y9ZZbzhUDXokvQ",
            cred_def_id="CsQY9MGeD3CQP4EyuVFo5m:3:CL:14951:MYCO_Biomarker",
            registry_type="CL_ACCUM",
            tag="tag",
            max_cred_num=100,
        )

        assert result is not None
        assert mock_handle.fetch.call_count == 1
        assert mock_tails_uri.call_count == 1
        assert mock_notify.call_count == 1

        # create doesn't fail with blank issuer id
        await self.revocation.create_and_register_revocation_registry_definition(
            issuer_id="",
            cred_def_id="CsQY9MGeD3CQP4EyuVFo5m:3:CL:14951:MYCO_Biomarker",
            registry_type="CL_ACCUM",
            tag="tag",
            max_cred_num=100,
        )

        # register registry response missing rev_reg_def_id and job_id
        self.profile.inject = mock.Mock(
            return_value=mock.MagicMock(
                register_revocation_registry_definition=mock.CoroutineMock(
                    side_effect=[
                        RevRegDefResult(
                            job_id=None,
                            revocation_registry_definition_state=RevRegDefState(
                                state=RevRegDefState.STATE_FINISHED,
                                revocation_registry_definition_id="active-reg-reg",
                                revocation_registry_definition=RevRegDef(
                                    tag="tag",
                                    cred_def_id="CsQY9MGeD3CQP4EyuVFo5m:3:CL:14951:MYCO_Biomarker",
                                    value=RevRegDefValue(
                                        max_cred_num=100,
                                        public_keys={
                                            "accum_key": {"z": "1 0BB...386"},
                                        },
                                        tails_hash="58NNWYnVxVFzAfUztwGSNBL4551XNq6nXk56pCiKJxxt",
                                        tails_location="http://tails-server.com",
                                    ),
                                    issuer_id="CsQY9MGeD3CQP4EyuVFo5m",
                                    type="CL_ACCUM",
                                ),
                            ),
                            registration_metadata={},
                            revocation_registry_definition_metadata={},
                        ),
                        RevRegDefResult(
                            job_id="test-job-id",
                            revocation_registry_definition_state=RevRegDefState(
                                state=RevRegDefState.STATE_FINISHED,
                                revocation_registry_definition_id=None,
                                revocation_registry_definition=RevRegDef(
                                    tag="tag",
                                    cred_def_id="CsQY9MGeD3CQP4EyuVFo5m:3:CL:14951:MYCO_Biomarker",
                                    value=RevRegDefValue(
                                        max_cred_num=100,
                                        public_keys={
                                            "accum_key": {"z": "1 0BB...386"},
                                        },
                                        tails_hash="58NNWYnVxVFzAfUztwGSNBL4551XNq6nXk56pCiKJxxt",
                                        tails_location="http://tails-server.com",
                                    ),
                                    issuer_id="CsQY9MGeD3CQP4EyuVFo5m",
                                    type="CL_ACCUM",
                                ),
                            ),
                            registration_metadata={},
                            revocation_registry_definition_metadata={},
                        ),
                        RevRegDefResult(
                            job_id=None,
                            revocation_registry_definition_state=RevRegDefState(
                                state=RevRegDefState.STATE_FINISHED,
                                revocation_registry_definition_id=None,
                                revocation_registry_definition=RevRegDef(
                                    tag="tag",
                                    cred_def_id="CsQY9MGeD3CQP4EyuVFo5m:3:CL:14951:MYCO_Biomarker",
                                    value=RevRegDefValue(
                                        max_cred_num=100,
                                        public_keys={
                                            "accum_key": {"z": "1 0BB...386"},
                                        },
                                        tails_hash="58NNWYnVxVFzAfUztwGSNBL4551XNq6nXk56pCiKJxxt",
                                        tails_location="http://tails-server.com",
                                    ),
                                    issuer_id="CsQY9MGeD3CQP4EyuVFo5m",
                                    type="CL_ACCUM",
                                ),
                            ),
                            registration_metadata={},
                            revocation_registry_definition_metadata={},
                        ),
                    ]
                )
            )
        )

        await self.revocation.create_and_register_revocation_registry_definition(
            issuer_id="did:indy:sovrin:SGrjRL82Y9ZZbzhUDXokvQ",
            cred_def_id="CsQY9MGeD3CQP4EyuVFo5m:3:CL:14951:MYCO_Biomarker",
            registry_type="CL_ACCUM",
            tag="tag",
            max_cred_num=100,
        )
        await self.revocation.create_and_register_revocation_registry_definition(
            issuer_id="did:indy:sovrin:SGrjRL82Y9ZZbzhUDXokvQ",
            cred_def_id="CsQY9MGeD3CQP4EyuVFo5m:3:CL:14951:MYCO_Biomarker",
            registry_type="CL_ACCUM",
            tag="tag",
            max_cred_num=100,
        )
        with self.assertRaises(test_module.AnonCredsRevocationError):
            await self.revocation.create_and_register_revocation_registry_definition(
                issuer_id="did:indy:sovrin:SGrjRL82Y9ZZbzhUDXokvQ",
                cred_def_id="CsQY9MGeD3CQP4EyuVFo5m:3:CL:14951:MYCO_Biomarker",
                registry_type="CL_ACCUM",
                tag="tag",
                max_cred_num=100,
            )

    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    @mock.patch.object(RevRegDef, "from_json", return_value="rev-reg-def")
    @mock.patch.object(test_module.AnonCredsRevocation, "notify")
    async def test_finish_revocation_registry_definition(
        self, mock_notify, mock_from_json, mock_handle
    ):
        mock_handle.fetch = mock.CoroutineMock(return_value=MockEntry())
        mock_handle.insert = mock.CoroutineMock(return_value=None)
        mock_handle.remove = mock.CoroutineMock(return_value=None)

        await self.revocation.finish_revocation_registry_definition(
            job_id="job-id",
            rev_reg_def_id="rev-reg-def-id",
        )

        # None response
        mock_handle.fetch = mock.CoroutineMock(return_value=None)
        with self.assertRaises(test_module.AnonCredsRevocationError):
            await self.revocation.finish_revocation_registry_definition(
                job_id="job-id",
                rev_reg_def_id="rev-reg-def-id",
            )

    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    async def test_get_created_revocation_registry_definitions(self, mock_handle):
        mock_handle.fetch_all = mock.CoroutineMock(
            return_value=[
                MockEntry("revocation_reg_def_0"),
                MockEntry("revocation_reg_def_1"),
            ]
        )
        result = await self.revocation.get_created_revocation_registry_definitions()
        assert result == ["revocation_reg_def_0", "revocation_reg_def_1"]

    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    async def test_get_created_revocation_registry_definition_state(self, mock_handle):
        mock_handle.fetch = mock.CoroutineMock(
            side_effect=[MockEntry(tags={"state": RevRegDefState.STATE_FINISHED}), None]
        )
        result = await self.revocation.get_created_revocation_registry_definition_state(
            "test-rev-reg-def-id"
        )
        assert result == RevRegDefState.STATE_FINISHED
        result = await self.revocation.get_created_revocation_registry_definition_state(
            "test-rev-reg-def-id"
        )
        assert result is None

    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    async def test_get_created_revocation_registry_definition(self, mock_handle):
        mock_handle.fetch = mock.CoroutineMock(
            side_effect=[
                MockEntry(
                    value_json=RevRegDef(
                        tag="tag",
                        cred_def_id="CsQY9MGeD3CQP4EyuVFo5m:3:CL:14951:MYCO_Biomarker",
                        value=RevRegDefValue(
                            max_cred_num=100,
                            public_keys={
                                "accum_key": {"z": "1 0BB...386"},
                            },
                            tails_hash="58NNWYnVxVFzAfUztwGSNBL4551XNq6nXk56pCiKJxxt",
                            tails_location="http://tails-server.com",
                        ),
                        issuer_id="CsQY9MGeD3CQP4EyuVFo5m",
                        type="CL_ACCUM",
                    ).to_json()
                ),
                None,
            ]
        )
        result = await self.revocation.get_created_revocation_registry_definition(
            "test-rev-reg-def-id"
        )
        assert isinstance(result, RevRegDef)
        result = await self.revocation.get_created_revocation_registry_definition(
            "test-rev-reg-def-id"
        )
        assert result is None

    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    async def test_set_active_registry(self, mock_handle):
        mock_handle.fetch = mock.CoroutineMock(return_value=None)
        mock_handle.replace = mock.CoroutineMock(return_value=None)
        inactive_tags = {
            "active": "false",
            "cred_def_id": "test-cred-def-id",
        }

        # fetch returns None
        with self.assertRaises(test_module.AnonCredsRevocationError):
            await self.revocation.set_active_registry(
                rev_reg_def_id="test-rev-reg-def-id",
            )
        # Already active
        mock_handle.fetch = mock.CoroutineMock(
            return_value=MockEntry(
                tags={
                    "active": "true",
                    "cred_def_id": "test-cred-def-id",
                }
            )
        )
        await self.revocation.set_active_registry(
            rev_reg_def_id="test-rev-reg-def-id",
        )

        mock_handle.fetch = mock.CoroutineMock(return_value=MockEntry(tags=inactive_tags))
        mock_handle.fetch_all = mock.CoroutineMock(
            return_value=[MockEntry(tags=inactive_tags), MockEntry(tags=inactive_tags)]
        )
        await self.revocation.set_active_registry(
            rev_reg_def_id="test-rev-reg-def-id",
        )

        assert mock_handle.fetch.call_count == 1
        assert mock_handle.fetch_all.call_count == 1
        assert mock_handle.replace.call_count == 3

    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    async def test_create_and_register_revocation_list_errors(self, mock_handle):
        class MockEntry:
            value_json = {
                "credDefId": "CsQY9MGeD3CQP4EyuVFo5m:3:CL:14951:MYCO_Biomarker",
            }

        mock_handle.fetch = mock.CoroutineMock(
            side_effect=[
                # failed to get cred def
                AskarError(code=AskarErrorCode.UNEXPECTED, message="test"),
                MockEntry(),
                # failed to get rev reg def
                MockEntry(),
                AskarError(code=AskarErrorCode.UNEXPECTED, message="test"),
                # failed to get cred def
                MockEntry(),
                MockEntry(),
                AskarError(code=AskarErrorCode.UNEXPECTED, message="test"),
            ]
        )
        # askar error
        for _ in range(3):
            with self.assertRaises(test_module.AnonCredsRevocationError):
                await self.revocation.create_and_register_revocation_list(
                    rev_reg_def_id="test-rev-reg-def-id",
                )

    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    @mock.patch.object(RevRegDef, "deserialize")
    @mock.patch.object(CredDef, "deserialize")
    @mock.patch.object(RevocationRegistryDefinitionPrivate, "load")
    @mock.patch.object(RevocationStatusList, "create")
    async def test_create_and_register_revocation_list(
        self,
        mock_list_create,
        mock_load_rev_list,
        mock_deserialize_cred_def,
        mock_deserialize_rev_reg,
        mock_handle,
    ):
        mock_list_create.return_value = RevocationStatusList.load(
            {
                "issuerId": "CsQY9MGeD3CQP4EyuVFo5m",
                "revRegDefId": "4xE68b6S5VRFrKMMG1U95M:4:4xE68b6S5VRFrKMMG1U95M:3:CL:59232:default:CL_ACCUM:4ae1cc6c-f6bd-486c-8057-88f2ce74e960",
                "revocationList": [0, 1, 1, 0],
                "currentAccumulator": "21 124C594B6B20E41B681E92B2C43FD165EA9E68BC3C9D63A82C8893124983CAE94 21 124C5341937827427B0A3A32113BD5E64FB7AB39BD3E5ABDD7970874501CA4897 6 5438CB6F442E2F807812FD9DC0C39AFF4A86B1E6766DBB5359E86A4D70401B0F 4 39D1CA5C4716FFC4FE0853C4FF7F081DFD8DF8D2C2CA79705211680AC77BF3A1 6 70504A5493F89C97C225B68310811A41AD9CD889301F238E93C95AD085E84191 4 39582252194D756D5D86D0EED02BF1B95CE12AED2FA5CD3C53260747D891993C",
                "timestamp": 1669640864487,
            }
        )
        mock_handle.fetch = mock.CoroutineMock(
            side_effect=[
                MockEntry(
                    value_json={
                        "credDefId": "CsQY9MGeD3CQP4EyuVFo5m:3:CL:14951:MYCO_Biomarker",
                        "issuerId": "CsQY9MGeD3CQP4EyuVFo5m",
                        "revocDefType": "CL_ACCUM",
                        "tag": "string",
                        "value": {
                            "maxCredNum": 0,
                            "publicKeys": {
                                "accumKey": {"z": "1 0BB...386"},
                            },
                            "tailsHash": "string",
                            "tailsLocation": "string",
                        },
                    }
                ),
                MockRevListEntry(),
                MockCredDefEntry(),
            ]
        )
        mock_handle.insert = mock.CoroutineMock(return_value=None)

        mock_registry = mock.MagicMock(AnonCredsRegistry, autospec=True)
        mock_registry.register_revocation_list = mock.CoroutineMock(
            return_value=RevListResult(
                job_id="test-job-id",
                revocation_list_state=RevListState(
                    revocation_list=rev_list,
                    state=RevListState.STATE_FINISHED,
                ),
                registration_metadata={},
                revocation_list_metadata={},
            )
        )
        self.profile.context.injector.bind_instance(AnonCredsRegistry, mock_registry)
        self.profile.context.injector.bind_instance(EventBus, MockEventBus())
        await self.revocation.create_and_register_revocation_list(
            rev_reg_def_id="test-rev-reg-def-id",
        )

        assert mock_handle.fetch.called
        assert mock_handle.insert.called
        assert mock_list_create.called
        assert mock_deserialize_cred_def.called
        assert mock_deserialize_rev_reg.called
        assert mock_load_rev_list.called
        assert self.profile.context.injector.get_provider(
            AnonCredsRegistry
        )._instance.register_revocation_list.called

    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    @mock.patch.object(test_module.AnonCredsRevocation, "_finish_registration")
    async def test_finish_revocation_list(self, mock_finish, mock_handle):
        self.profile.context.injector.bind_instance(EventBus, MockEventBus())

        mock_handle.fetch = mock.CoroutineMock(side_effect=[None, MockEntry()])

        # Fetch doesn't find list then it should be created
        await self.revocation.finish_revocation_list(
            job_id="test-job-id", rev_reg_def_id="test-rev-reg-def-id", revoked=[]
        )
        assert mock_finish.called

        # Fetch finds list then there's nothing to do, it's already finished and updated
        await self.revocation.finish_revocation_list(
            job_id="test-job-id", rev_reg_def_id="test-rev-reg-def-id", revoked=[]
        )
        assert mock_finish.call_count == 1

    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    async def test_update_revocation_list_get_rev_reg_errors(self, mock_handle):
        mock_handle.fetch = mock.CoroutineMock(
            side_effect=[
                AskarError(code=AskarErrorCode.UNEXPECTED, message="test"),
                None,
            ]
        )
        # askar error
        with self.assertRaises(test_module.AnonCredsRevocationError):
            await self.revocation.update_revocation_list(
                rev_reg_def_id="test-rev-reg-def-id",
                prev=rev_list,
                curr=rev_list,
                revoked=[1, 1, 0, 0],
            )
        # fetch returns None
        with self.assertRaises(test_module.AnonCredsRevocationError):
            await self.revocation.update_revocation_list(
                rev_reg_def_id="test-rev-reg-def-id",
                prev=rev_list,
                curr=rev_list,
                revoked=[1, 1, 0, 0],
            )

    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    async def test_update_revocation_list(self, mock_handle):
        mock_handle.fetch = mock.CoroutineMock(
            side_effect=[
                MockRevRegDefEntry(),
                MockRevListEntry(),
                MockRevListEntry(),
            ]
        )
        mock_handle.replace = mock.CoroutineMock(return_value=None)

        self.profile.inject = mock.Mock(
            return_value=mock.MagicMock(
                update_revocation_list=mock.CoroutineMock(
                    return_value=RevListResult(
                        job_id="test-job-id",
                        revocation_list_state=RevListState(
                            revocation_list=rev_list,
                            state=RevListState.STATE_FINISHED,
                        ),
                        registration_metadata={},
                        revocation_list_metadata={},
                    )
                )
            )
        )

        # valid with no errors
        result = await self.revocation.update_revocation_list(
            rev_reg_def_id="test-rev-reg-def-id",
            prev=rev_list,
            curr=rev_list,
            revoked=[1, 1, 0, 0],
        )

        assert mock_handle.fetch.call_count == 3
        assert mock_handle.replace.called
        assert result is not None

        # askar error fetching list is caught
        mock_handle.fetch = mock.CoroutineMock(
            side_effect=[
                MockRevRegDefEntry(),
                AskarError(code=AskarErrorCode.UNEXPECTED, message="test"),
            ]
        )
        with self.assertRaises(test_module.AnonCredsRevocationError):
            await self.revocation.update_revocation_list(
                rev_reg_def_id="test-rev-reg-def-id",
                prev=rev_list,
                curr=rev_list,
                revoked=[1, 1, 0, 0],
            )

        # fail to get list
        mock_handle.fetch = mock.CoroutineMock(
            side_effect=[
                MockRevRegDefEntry(),
                None,
            ]
        )
        with self.assertRaises(test_module.AnonCredsRevocationError):
            await self.revocation.update_revocation_list(
                rev_reg_def_id="test-rev-reg-def-id",
                prev=rev_list,
                curr=rev_list,
                revoked=[1, 1, 0, 0],
            )

        # revocation lists don't match
        mock_handle.fetch = mock.CoroutineMock(
            side_effect=[
                MockRevRegDefEntry(),
                MockRevListEntry(),
            ]
        )
        with self.assertRaises(test_module.AnonCredsRevocationError):
            await self.revocation.update_revocation_list(
                rev_reg_def_id="test-rev-reg-def-id",
                prev=rev_list,
                curr=RevList(
                    issuer_id="CsQY9MGeD3CQP4EyuVFo5m",
                    current_accumulator="21 124C594B6B20E41B681E92B2C43FD165EA9E68BC3C9D63A82C8893124983CAE94 21 124C5341937827427B0A3A32113BD5E64FB7AB39BD3E5ABDD7970874501CA4897 6 5438CB6F442E2F807812FD9DC0C39AFF4A86B1E6766DBB5359E86A4D70401B0F 4 39D1CA5C4716FFC4FE0853C4FF7F081DFD8DF8D2C2CA79705211680AC77BF3A1 6 70504A5493F89C97C225B68310811A41AD9CD889301F238E93C95AD085E84191 4 39582252194D756D5D86D0EED02BF1B95CE12AED2FA5CD3C53260747D891993C",
                    revocation_list=[1, 0, 1, 0],
                    timestamp=1669640864487,
                    rev_reg_def_id="4xE68b6S5VRFrKMMG1U95M:4:4xE68b6S5VRFrKMMG1U95M:3:CL:59232:default:CL_ACCUM:4ae1cc6c-f6bd-486c-8057-88f2ce74e960",
                ),
                revoked=[1, 1, 0, 0],
            )

        # update fail states are caught
        mock_handle.fetch = mock.CoroutineMock(
            side_effect=[
                MockRevRegDefEntry(),
                MockRevListEntry(),
                AskarError(code=AskarErrorCode.UNEXPECTED, message="test"),
            ]
        )
        with self.assertRaises(test_module.AnonCredsRevocationError):
            await self.revocation.update_revocation_list(
                rev_reg_def_id="test-rev-reg-def-id",
                prev=rev_list,
                curr=rev_list,
                revoked=[1, 1, 0, 0],
            )
        mock_handle.fetch = mock.CoroutineMock(
            side_effect=[
                MockRevRegDefEntry(),
                MockRevListEntry(),
                None,
            ]
        )
        with self.assertRaises(test_module.AnonCredsRevocationError):
            await self.revocation.update_revocation_list(
                rev_reg_def_id="test-rev-reg-def-id",
                prev=rev_list,
                curr=rev_list,
                revoked=[1, 1, 0, 0],
            )

    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    async def test_get_created_revocation_list(self, mock_handle):
        mock_handle.fetch = mock.CoroutineMock(
            side_effect=[
                MockRevListEntry(),
                None,
                AskarError(code=AskarErrorCode.UNEXPECTED, message="test"),
            ]
        )
        result = await self.revocation.get_created_revocation_list("rev-reg-def-id")
        assert mock_handle.fetch.call_count == 1
        assert result is not None

        # fetch returns None
        result = await self.revocation.get_created_revocation_list("rev-reg-def-id")
        assert result is None

        # askar error
        with self.assertRaises(test_module.AnonCredsRevocationError):
            await self.revocation.get_created_revocation_list("rev-reg-def-id")

    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    async def test_get_revocation_lists_with_pending_revocations(self, mock_handle):
        mock_handle.fetch_all = mock.CoroutineMock(
            side_effect=[
                [MockEntry("rev_list_0"), MockEntry("rev_list_1")],
                [],
                AskarError(code=AskarErrorCode.UNEXPECTED, message="test"),
            ]
        )
        result = await self.revocation.get_revocation_lists_with_pending_revocations()
        assert result == ["rev_list_0", "rev_list_1"]

        # fetch returns None
        result = await self.revocation.get_revocation_lists_with_pending_revocations()
        assert result == []

        # askar error
        with self.assertRaises(test_module.AnonCredsRevocationError):
            await self.revocation.get_revocation_lists_with_pending_revocations()

    @mock.patch.object(Session, "get")
    @mock.patch.object(os, "remove")
    async def test_retrieve_tails(self, mock_remove, mock_get):
        class MockResponse:
            def __init__(self, status=http.HTTPStatus.OK):
                self.status_code = status

            def iter_content(self, chunk_size: int = 1):
                yield b"tails-hash"

        mock_get.side_effect = [
            MockResponse(),
            MockResponse(),
            RequestException(request=mock.AsyncMock(), response=mock.AsyncMock()),
            MockResponse(status=http.HTTPStatus.BAD_REQUEST),
        ]

        result = await self.revocation.retrieve_tails(rev_reg_def)

        assert isinstance(result, str)
        assert mock_get.call_count == 1

        # tails hash does not match
        with self.assertRaises(test_module.AnonCredsRevocationError):
            await self.revocation.retrieve_tails(
                RevRegDef(
                    tag="tag",
                    cred_def_id="CsQY9MGeD3CQP4EyuVFo5m:3:CL:14951:MYCO_Biomarker",
                    value=RevRegDefValue(
                        max_cred_num=100,
                        public_keys={
                            "accum_key": {"z": "1 0BB...386"},
                        },
                        tails_hash="not-correct-hash",
                        tails_location="http://tails-server.com",
                    ),
                    issuer_id="CsQY9MGeD3CQP4EyuVFo5m",
                    type="CL_ACCUM",
                )
            )
            assert mock_remove.call_count == 1

        # http request fails
        with self.assertRaises(test_module.AnonCredsRevocationError):
            await self.revocation.retrieve_tails(rev_reg_def)

        # doesn't crash on non 200 response
        await self.revocation.retrieve_tails(rev_reg_def)

    def test_generate_public_tails_uri(self):
        self.revocation.generate_public_tails_uri(rev_reg_def)

        # invalid url
        self.profile.settings["tails_server_base_url"] = "invalid-url"
        with self.assertRaises(test_module.AnonCredsRevocationError):
            self.revocation.generate_public_tails_uri(rev_reg_def)

        # tails server base url setting is missing
        del self.profile.settings["tails_server_base_url"]
        with self.assertRaises(test_module.AnonCredsRevocationError):
            self.revocation.generate_public_tails_uri(rev_reg_def)

    async def test_upload_tails_file(self):
        tails_server = mock.MagicMock(AnonCredsTailsServer, autospec=True)
        tails_server.upload_tails_file = mock.CoroutineMock(
            return_value=(True, "https://tails-server.com")
        )

        # valid
        with mock.patch.object(
            test_module,
            "AnonCredsTailsServer",
            mock.MagicMock(return_value=tails_server),
        ):
            await self.revocation.upload_tails_file(rev_reg_def)
        tails_server.upload_tails_file.assert_called_once()

        # upload fails
        tails_server.upload_tails_file.reset_mock()
        tails_server.upload_tails_file = mock.CoroutineMock(
            return_value=(None, "https://tails-server.com")
        )
        with (
            mock.patch.object(
                test_module,
                "AnonCredsTailsServer",
                mock.MagicMock(return_value=tails_server),
            ),
            self.assertRaises(test_module.AnonCredsRevocationError),
        ):
            await self.revocation.upload_tails_file(rev_reg_def)

        tails_server.upload_tails_file.reset_mock()
        tails_server.upload_tails_file = mock.CoroutineMock(
            return_value=(None, "not-https://tails-server.com")
        )

        # tails location does not match
        with (
            mock.patch.object(
                test_module,
                "AnonCredsTailsServer",
                mock.MagicMock(return_value=tails_server),
            ),
            self.assertRaises(test_module.AnonCredsRevocationError),
        ):
            await self.revocation.upload_tails_file(rev_reg_def)

    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    @mock.patch.object(
        test_module.AnonCredsRevocation, "set_active_registry", return_value=None
    )
    @mock.patch.object(
        test_module.AnonCredsRevocation,
        "create_and_register_revocation_registry_definition",
        return_value=MockRevRegDefEntry(),
    )
    async def test_handle_full_registry(
        self, mock_create_and_register, mock_set_active_registry, mock_handle
    ):
        mock_handle.fetch = mock.CoroutineMock(return_value=MockRevRegDefEntry())
        mock_handle.fetch_all = mock.CoroutineMock(
            return_value=[
                MockRevRegDefEntry(),
                MockRevRegDefEntry(),
            ]
        )
        mock_handle.replace = mock.CoroutineMock(return_value=None)

        await self.revocation.handle_full_registry("test-rev-reg-def-id")
        assert mock_create_and_register.called
        assert mock_set_active_registry.called
        assert mock_handle.fetch.call_count == 2
        assert mock_handle.fetch_all.called
        assert mock_handle.replace.called

        # no backup registry available
        mock_handle.fetch_all = mock.CoroutineMock(return_value=[])
        with self.assertRaises(test_module.AnonCredsRevocationError):
            await self.revocation.handle_full_registry("test-rev-reg-def-id")

    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    async def test_decommission_registry(self, mock_handle):
        mock_handle.fetch_all = mock.CoroutineMock(
            return_value=[
                MockEntry(
                    name="active-reg-reg",
                    tags={
                        "state": RevRegDefState.STATE_FINISHED,
                        "active": True,
                    },
                ),
                MockEntry(
                    name="new-rev-reg",
                    tags={
                        "state": RevRegDefState.STATE_FINISHED,
                        "active": True,
                    },
                ),
            ]
        )
        # active registry
        self.revocation.get_or_create_active_registry = mock.CoroutineMock(
            return_value=RevRegDefResult(
                job_id="test-job-id",
                revocation_registry_definition_state=RevRegDefState(
                    state=RevRegDefState.STATE_FINISHED,
                    revocation_registry_definition_id="active-reg-reg",
                    revocation_registry_definition=rev_reg_def,
                ),
                registration_metadata={},
                revocation_registry_definition_metadata={},
            )
        )
        # new active
        self.revocation.create_and_register_revocation_registry_definition = (
            mock.CoroutineMock(
                return_value=RevRegDefResult(
                    job_id="test-job-id",
                    revocation_registry_definition_state=RevRegDefState(
                        state=RevRegDefState.STATE_ACTION,
                        revocation_registry_definition_id="new-rev-reg",
                        revocation_registry_definition=rev_reg_def,
                    ),
                    registration_metadata={},
                    revocation_registry_definition_metadata={},
                )
            )
        )
        self.revocation.set_active_registry = mock.CoroutineMock(return_value=None)
        mock_handle.replace = mock.CoroutineMock(return_value=None)

        result = await self.revocation.decommission_registry("test-rev-reg-def-id")

        assert isinstance(result, list)
        assert len(result) == 2
        assert result[0].tags["active"] == "false"
        assert result[0].tags["state"] == RevRegDefState.STATE_DECOMMISSIONED
        assert mock_handle.fetch_all.called
        assert mock_handle.replace.called
        # # One for backup
        assert (
            self.revocation.create_and_register_revocation_registry_definition.call_count
            == 2
        )

    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    async def test_get_or_create_active_registry(self, mock_handle):
        mock_handle.fetch_all = mock.CoroutineMock(
            side_effect=[
                [MockRevRegDefEntry("reg-1"), MockRevRegDefEntry("reg-0")],
                None,
            ]
        )

        # valid
        result = await self.revocation.get_or_create_active_registry(
            "test-rev-reg-def-id"
        )
        assert isinstance(result, RevRegDefResult)
        assert result.revocation_registry_definition_state.state == (
            RevRegDefState.STATE_FINISHED
        )
        assert (
            result.revocation_registry_definition_state.revocation_registry_definition_id
            == ("reg-1")
        )

        # no active registry, todo: create one
        with self.assertRaises(test_module.AnonCredsRevocationError):
            await self.revocation.get_or_create_active_registry("test-rev-reg-def-id")

    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    @mock.patch.object(Credential, "create", return_value=mock.MagicMock())
    async def test_create_credential_private_no_rev_reg_or_tails(
        self, mock_create, mock_handle
    ):
        mock_handle.fetch = mock.CoroutineMock(side_effect=[MockEntry(), MockEntry()])
        await self.revocation._create_credential(
            credential_definition_id="test-cred-def-id",
            schema_attributes=["attr1", "attr2"],
            credential_offer={
                "schema_id": "CsQY9MGeD3CQP4EyuVFo5m:2:MYCO Biomarker:0.0.3",
                "cred_def_id": "CsQY9MGeD3CQP4EyuVFo5m:3:CL:14951:MYCO_Biomarker",
                "key_correctness_proof": {},
                "nonce": "nonce",
            },
            credential_request={},
            credential_values={
                "attr1": "value1",
                "attr2": "value2",
            },
            credential_type=Credential,
        )
        assert mock_create.called

        # askar error retrieving cred def
        mock_handle.fetch = mock.CoroutineMock(
            side_effect=AskarError(AskarErrorCode.UNEXPECTED, "test")
        )
        with self.assertRaises(test_module.AnonCredsRevocationError):
            await self.revocation._create_credential(
                credential_definition_id="test-cred-def-id",
                schema_attributes=["attr1", "attr2"],
                credential_offer={},
                credential_request={},
                credential_values={},
                credential_type=Credential,
            )

        # missing cred def or cred def private
        mock_handle.fetch = mock.CoroutineMock(side_effect=[None, MockEntry()])
        with self.assertRaises(test_module.AnonCredsRevocationError):
            await self.revocation._create_credential(
                credential_definition_id="test-cred-def-id",
                schema_attributes=["attr1", "attr2"],
                credential_offer={},
                credential_request={},
                credential_values={},
                credential_type=Credential,
            )
        mock_handle.fetch = mock.CoroutineMock(side_effect=[MockEntry(), None])
        with self.assertRaises(test_module.AnonCredsRevocationError):
            await self.revocation._create_credential(
                credential_definition_id="test-cred-def-id",
                schema_attributes=["attr1", "attr2"],
                credential_offer={},
                credential_request={},
                credential_values={},
                credential_type=Credential,
            )

    @mock.patch.object(
        RevocationRegistryDefinition, "load", return_value=rev_reg_def.value
    )
    @mock.patch("acapy_agent.anoncreds.revocation.CredentialRevocationConfig")
    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    @mock.patch.object(Credential, "create", return_value=mock.MagicMock())
    async def test_create_credential_private_with_rev_reg_and_tails(
        self, mock_create, mock_handle, *_
    ):
        async def call_test_func():
            await self.revocation._create_credential(
                credential_definition_id="test-cred-def-id",
                schema_attributes=["attr1", "attr2"],
                credential_offer={
                    "schema_id": "CsQY9MGeD3CQP4EyuVFo5m:2:MYCO Biomarker:0.0.3",
                    "cred_def_id": "CsQY9MGeD3CQP4EyuVFo5m:3:CL:14951:MYCO_Biomarker",
                    "key_correctness_proof": {},
                    "nonce": "nonce",
                },
                credential_request={},
                credential_values={
                    "attr1": "value1",
                    "attr2": "value2",
                },
                rev_reg_def_id="test-rev-reg-def-id",
                tails_file_path="tails-file-path",
                credential_type=Credential,
            )

        # missing rev def
        mock_handle.fetch = mock.CoroutineMock(
            side_effect=[None, MockEntry(), MockEntry()]
        )
        with self.assertRaises(test_module.AnonCredsRevocationError):
            await call_test_func()
        # missing rev list
        mock_handle.fetch = mock.CoroutineMock(
            side_effect=[MockEntry(), None, MockEntry()]
        )
        with self.assertRaises(test_module.AnonCredsRevocationError):
            await call_test_func()
        # missing rev key
        mock_handle.fetch = mock.CoroutineMock(
            side_effect=[MockEntry(), MockEntry(), None]
        )
        with self.assertRaises(test_module.AnonCredsRevocationError):
            await call_test_func()

        # valid
        mock_handle.replace = mock.CoroutineMock(return_value=None)
        mock_handle.fetch = mock.CoroutineMock(
            side_effect=[
                MockEntry(raw_value=rev_reg_def.serialize()),
                MockEntry(
                    value_json={
                        "rev_list": rev_list.serialize(),
                        "next_index": 0,
                    }
                ),
                MockEntry(),
                MockEntry(),
                MockEntry(),
            ]
        )
        await call_test_func()
        assert mock_create.called
        assert mock_handle.replace.called

        # revocation registry is full
        mock_handle.fetch = mock.CoroutineMock(
            side_effect=[
                MockEntry(raw_value=rev_reg_def.serialize()),
                MockEntry(
                    value_json={
                        "rev_list": rev_list.serialize(),
                        "next_index": 101,
                    }
                ),
                MockEntry(),
                MockEntry(),
                MockEntry(),
            ]
        )
        with self.assertRaises(test_module.AnonCredsRevocationError):
            await call_test_func()

    @mock.patch.object(AnonCredsIssuer, "cred_def_supports_revocation", return_value=True)
    async def test_create_credential(self, mock_supports_revocation):
        self.profile.inject = mock.Mock(
            return_value=mock.MagicMock(
                get_schema=mock.CoroutineMock(
                    return_value=GetSchemaResult(
                        schema_id="CsQY9MGeD3CQP4EyuVFo5m:2:MYCO Biomarker:0.0.3",
                        schema=AnonCredsSchema(
                            issuer_id="CsQY9MGeD3CQP4EyuVFo5m",
                            name="MYCO Biomarker:0.0.3",
                            version="1.0",
                            attr_names=["attr1", "attr2"],
                        ),
                        schema_metadata={},
                        resolution_metadata={},
                    )
                )
            )
        )
        self.revocation.get_or_create_active_registry = mock.CoroutineMock(
            return_value=RevRegDefResult(
                job_id="test-job-id",
                revocation_registry_definition_state=RevRegDefState(
                    state=RevRegDefState.STATE_FINISHED,
                    revocation_registry_definition_id="active-reg-reg",
                    revocation_registry_definition=rev_reg_def,
                ),
                registration_metadata={},
                revocation_registry_definition_metadata={},
            )
        )

        # Test private funtion seperately - very large
        self.revocation._create_credential = mock.CoroutineMock(
            return_value=({"cred": "cred"}, 98)
        )

        result = await self.revocation.create_credential(
            credential_offer={
                "schema_id": "CsQY9MGeD3CQP4EyuVFo5m:2:MYCO Biomarker:0.0.3",
                "cred_def_id": "CsQY9MGeD3CQP4EyuVFo5m:3:CL:14951:MYCO_Biomarker",
                "key_correctness_proof": {},
                "nonce": "nonce",
            },
            credential_request={},
            credential_values={},
        )

        assert isinstance(result, tuple)
        assert mock_supports_revocation.call_count == 1

    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    @mock.patch.object(RevList, "to_native")
    @mock.patch.object(RevList, "from_native", return_value=None)
    @mock.patch.object(RevRegDef, "to_native")
    @mock.patch.object(CredDef, "deserialize")
    @mock.patch.object(RevocationRegistryDefinitionPrivate, "load")
    async def test_revoke_pending_credentials(
        self,
        mock_load_rev_reg,
        mock_deserialize_cred_def,
        mock_rev_reg_to_native,
        mock_rev_list_from_native,
        mock_rev_list_to_native,
        mock_handle,
    ):
        mock_handle.fetch = mock.CoroutineMock(
            side_effect=[
                # askar error
                AskarError(code=AskarErrorCode.UNEXPECTED, message="test"),
                # missing rev reg def
                None,
                MockEntry(value_json="{}"),
                MockEntry(value_json="{}"),
                # missing rev list
                MockEntry(value_json="{}"),
                None,
                MockEntry(value_json="{}"),
                # missing rev private
                MockEntry(value_json="{}"),
                MockEntry(value_json="{}"),
                None,
            ]
        )

        # askar error
        with self.assertRaises(test_module.AnonCredsRevocationError):
            await self.revocation.revoke_pending_credentials(
                revoc_reg_id="test-rev-reg-id",
            )
        # rev reg def not found
        with self.assertRaises(test_module.AnonCredsRevocationError):
            await self.revocation.revoke_pending_credentials(
                revoc_reg_id="test-rev-reg-id",
            )
        # rev red def private not found
        with self.assertRaises(test_module.AnonCredsRevocationError):
            await self.revocation.revoke_pending_credentials(
                revoc_reg_id="test-rev-reg-id",
            )
        # rev list not found
        with self.assertRaises(test_module.AnonCredsRevocationError):
            await self.revocation.revoke_pending_credentials(
                revoc_reg_id="test-rev-reg-id",
            )

        # valid
        mock_handle.fetch = mock.CoroutineMock(
            side_effect=[
                # rev_reg_def_entry
                MockEntry(value_json=MOCK_REV_REG_DEF),
                # rev_list_entry
                MockEntry(
                    value_json={
                        "pending": [0, 1, 4, 3],
                        "next_index": 4,
                        "rev_list": rev_list.serialize(),
                    }
                ),
                # private rev reg def
                MockEntry(),
                # cred def
                MockEntry(),
                # updated rev list entry
                MockEntry(
                    value_json={
                        "pending": [0, 1, 4, 3],
                        "next_index": 4,
                        "rev_list": rev_list.serialize(),
                    },
                    tags={"pending": []},
                ),
            ]
        )
        mock_handle.replace = mock.CoroutineMock(return_value=None)

        result = await self.revocation.revoke_pending_credentials(
            revoc_reg_id="test-rev-reg-id",
        )

        assert mock_handle.fetch.call_count == 5
        assert mock_handle.replace.called
        assert mock_rev_list_from_native.called
        assert mock_rev_list_to_native.called
        assert mock_rev_reg_to_native.called
        assert mock_load_rev_reg.called
        assert mock_deserialize_cred_def.called
        assert isinstance(result, test_module.RevokeResult)

    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    @mock.patch.object(RevList, "to_native")
    @mock.patch.object(RevList, "from_native", return_value=None)
    @mock.patch.object(RevRegDef, "to_native")
    @mock.patch.object(CredDef, "deserialize")
    @mock.patch.object(RevocationRegistryDefinitionPrivate, "load")
    async def test_revoke_pending_credentials_cred_def_error(
        self,
        mock_load_rev_reg,
        mock_deserialize_cred_def,
        mock_rev_reg_to_native,
        mock_rev_list_from_native,
        mock_rev_list_to_native,
        mock_handle,
    ):
        """Test error handling when fetching credential definition fails."""
        mock_handle.fetch = mock.CoroutineMock(
            side_effect=[
                MockEntry(value_json=MOCK_REV_REG_DEF),
                MockEntry(
                    value_json={
                        "pending": [0, 1, 4, 3],
                        "next_index": 4,
                        "rev_list": rev_list.serialize(),
                    }
                ),
                MockEntry(),
                AskarError(code=AskarErrorCode.UNEXPECTED, message="test error"),
            ]
        )

        with self.assertRaises(test_module.AnonCredsRevocationError) as cm:
            await self.revocation.revoke_pending_credentials(
                revoc_reg_id="test-rev-reg-id",
            )
        assert "Error retrieving cred def" in str(cm.exception)

    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    @mock.patch.object(RevList, "to_native")
    @mock.patch.object(RevList, "from_native", return_value=None)
    @mock.patch.object(RevRegDef, "to_native")
    @mock.patch.object(CredDef, "deserialize")
    @mock.patch.object(RevocationRegistryDefinitionPrivate, "load")
    async def test_revoke_pending_credentials_anoncreds_error(
        self,
        mock_load_rev_reg,
        mock_deserialize_cred_def,
        mock_rev_reg_to_native,
        mock_rev_list_from_native,
        mock_rev_list_to_native,
        mock_handle,
    ):
        """Test error handling when loading revocation registry definition fails."""
        mock_handle.fetch = mock.CoroutineMock(
            side_effect=[
                MockEntry(value_json=MOCK_REV_REG_DEF),
                MockEntry(
                    value_json={
                        "pending": [0, 1, 4, 3],
                        "next_index": 4,
                        "rev_list": rev_list.serialize(),
                    }
                ),
                MockEntry(),
                MockEntry(),
            ]
        )
        mock_deserialize_cred_def.side_effect = AnoncredsError(
            AnoncredsErrorCode.UNEXPECTED, "Failed to load"
        )

        with self.assertRaises(test_module.AnonCredsRevocationError) as cm:
            await self.revocation.revoke_pending_credentials(
                revoc_reg_id="test-rev-reg-id",
            )
        assert "Error loading revocation registry definition" in str(cm.exception)

    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    @mock.patch.object(RevList, "to_native")
    @mock.patch.object(RevList, "from_native", return_value=None)
    @mock.patch.object(RevRegDef, "to_native")
    @mock.patch.object(CredDef, "deserialize")
    @mock.patch.object(RevocationRegistryDefinitionPrivate, "load")
    async def test_revoke_pending_credentials_no_valid_credentials(
        self,
        mock_load_rev_reg,
        mock_deserialize_cred_def,
        mock_rev_reg_to_native,
        mock_rev_list_from_native,
        mock_rev_list_to_native,
        mock_handle,
    ):
        """Test handling when there are no valid credentials to revoke."""
        mock_handle.fetch = mock.CoroutineMock(
            side_effect=[
                MockEntry(value_json=MOCK_REV_REG_DEF),
                MockEntry(
                    value_json={
                        "pending": [],  # No pending revocations
                        "next_index": 4,
                        "rev_list": rev_list.serialize(),
                    }
                ),
                MockEntry(),
                MockEntry(),
            ]
        )

        result = await self.revocation.revoke_pending_credentials(
            revoc_reg_id="test-rev-reg-id",
        )
        assert result.revoked == []  # No credentials were revoked
        assert result.failed == []  # No failures

    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    async def test_mark_pending_revocations(self, mock_handle):
        mock_handle.fetch = mock.CoroutineMock(
            side_effect=[
                None,
                MockEntry(
                    value_json={
                        "pending": [1],
                    }
                ),
            ]
        )
        mock_handle.replace = mock.CoroutineMock(return_value=None)

        # rev list entry not found
        with self.assertRaises(test_module.AnonCredsRevocationError):
            await self.revocation.mark_pending_revocations("test-rev-reg-id", int("200"))

        # valid
        await self.revocation.mark_pending_revocations("test-rev-reg-id", int("200"))
        assert mock_handle.replace.call_count == 1

    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    async def test_get_pending_revocations(self, mock_handle):
        mock_handle.fetch = mock.CoroutineMock(
            side_effect=[
                None,
                MockEntry(
                    value_json={
                        "pending": [1, 2],
                    }
                ),
            ]
        )

        result = await self.revocation.get_pending_revocations("test-rev-reg-id")
        assert result == []

        result = await self.revocation.get_pending_revocations("test-rev-reg-id")
        assert result == [1, 2]

    @mock.patch("acapy_agent.anoncreds.revocation.isinstance")
    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    async def test_clear_pending_revocations(self, mock_handle, mock_is_instance):
        mock_handle.fetch = mock.CoroutineMock(
            side_effect=[
                None,
                MockEntry(
                    value_json={
                        "pending": [1, 2],
                    }
                ),
                MockEntry(
                    value_json={
                        "pending": [1, 2],
                    }
                ),
            ]
        )
        mock_handle.replace = mock.CoroutineMock(return_value=None)

        # fetch is None
        async with self.profile.session() as session:
            with self.assertRaises(test_module.AnonCredsRevocationError):
                await self.revocation.clear_pending_revocations(
                    session, rev_reg_def_id="test-rev-reg-id"
                )
            # valid
            await self.revocation.clear_pending_revocations(
                session, rev_reg_def_id="test-rev-reg-id"
            )
            assert mock_handle.replace.called
            # with crid mask
            await self.revocation.clear_pending_revocations(
                session, rev_reg_def_id="test-rev-reg-id", crid_mask=[1, 2]
            )

    async def test_clear_pending_revocations_with_non_anoncreds_session(self):
        self.profile = await create_test_profile(settings={"wallet.type": "askar"})
        with self.assertRaises(ValueError):
            await self.revocation.clear_pending_revocations(
                self.profile.session(), rev_reg_def_id="test-rev-reg-id"
            )

    @mock.patch.object(AnonCredsIssuer, "cred_def_supports_revocation", return_value=True)
    async def test_create_credential_w3c(self, mock_supports_revocation):
        self.profile.inject = mock.Mock(
            return_value=mock.MagicMock(
                get_schema=mock.CoroutineMock(
                    return_value=GetSchemaResult(
                        schema_id="CsQY9MGeD3CQP4EyuVFo5m:2:MYCO Biomarker:0.0.3",
                        schema=AnonCredsSchema(
                            issuer_id="CsQY9MGeD3CQP4EyuVFo5m",
                            name="MYCO Biomarker:0.0.3",
                            version="1.0",
                            attr_names=["attr1", "attr2"],
                        ),
                        schema_metadata={},
                        resolution_metadata={},
                    )
                )
            )
        )
        self.revocation.get_or_create_active_registry = mock.CoroutineMock(
            return_value=RevRegDefResult(
                job_id="test-job-id",
                revocation_registry_definition_state=RevRegDefState(
                    state=RevRegDefState.STATE_FINISHED,
                    revocation_registry_definition_id="active-reg-reg",
                    revocation_registry_definition=rev_reg_def,
                ),
                registration_metadata={},
                revocation_registry_definition_metadata={},
            )
        )

        # Test private funtion seperately - very large
        self.revocation._create_credential = mock.CoroutineMock(
            return_value=({"cred": "cred"}, 98)
        )

        result = await self.revocation.create_credential_w3c(
            w3c_credential_offer={
                "schema_id": "CsQY9MGeD3CQP4EyuVFo5m:2:MYCO Biomarker:0.0.3",
                "cred_def_id": "CsQY9MGeD3CQP4EyuVFo5m:3:CL:14951:MYCO_Biomarker",
                "key_correctness_proof": {},
                "nonce": "nonce",
            },
            w3c_credential_request={},
            w3c_credential_values={},
        )

        assert isinstance(result, tuple)
        assert mock_supports_revocation.call_count == 1

    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    async def test_create_credential_w3c_keyerror(self, mock_handle):
        mock_handle.fetch = mock.CoroutineMock(side_effect=[MockEntry(), MockEntry()])
        with pytest.raises(test_module.AnonCredsRevocationError) as excinfo:
            await self.revocation._create_credential(
                credential_definition_id="test-cred-def-id",
                schema_attributes=["attr1", "attr2"],
                credential_offer={
                    "schema_id": "CsQY9MGeD3CQP4EyuVFo5m:2:MYCO Biomarker:0.0.3",
                    "cred_def_id": "CsQY9MGeD3CQP4EyuVFo5m:3:CL:14951:MYCO_Biomarker",
                    "key_correctness_proof": {},
                    "nonce": "nonce",
                },
                credential_request={},
                credential_values={
                    "X": "value1",
                    "Y": "value2",
                },
                credential_type=Credential,
            )

        assert str(excinfo.value) == (
            "Provided credential values are missing a value "
            "for the schema attribute 'attr1'"
        )



from typing import Tuple
from unittest import IsolatedAsyncioTestCase

import pytest

from ...resolver.did_resolver import DIDResolver
from ...resolver.tests.test_did_resolver import MockResolver
from ...utils.testing import create_test_profile
from ...wallet.did_method import KEY, DIDMethods
from ...wallet.key_type import ED25519, P256, KeyType, KeyTypes
from ..base import BaseWallet
from ..default_verification_key_strategy import (
    BaseVerificationKeyStrategy,
    DefaultVerificationKeyStrategy,
)
from ..jwt import jwt_sign, jwt_verify, resolve_public_key_by_kid_for_verify


class TestJWT(IsolatedAsyncioTestCase):
    """Tests for JWT sign and verify using dids."""

    seed = "testseed000000000000000000000001"
    did_key_ed25519_did = "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"
    did_key_ed25519_verification_method = "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"
    did_key_ed25519_doc = {
        "@context": "https://www.w3.org/ns/did/v1",
        "id": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
        "verificationMethod": [
            {
                "id": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                "type": "Ed25519VerificationKey2018",
                "controller": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                "publicKeyBase58": "3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRx",
            }
        ],
        "authentication": [
            "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"
        ],
        "assertionMethod": [
            "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"
        ],
        "capabilityDelegation": [
            "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"
        ],
        "capabilityInvocation": [
            "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"
        ],
        "keyAgreement": [
            {
                "id": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6LSbkodSr6SU2trs8VUgnrnWtSm7BAPG245ggrBmSrxbv1R",
                "type": "X25519KeyAgreementKey2019",
                "controller": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                "publicKeyBase58": "5dTvYHaNaB7mk7iA9LqCJEHG2dGZQsvoi8WGzDRtYEf",
            }
        ],
    }
    did_key_p256_did = "did:key:zDnaehWviigWQQD7bqF3btFquwA5w8DX2sQwkVxnAyJ7oxdjq"
    did_key_p256_verification_method = "did:key:zDnaehWviigWQQD7bqF3btFquwA5w8DX2sQwkVxnAyJ7oxdjq#zDnaehWviigWQQD7bqF3btFquwA5w8DX2sQwkVxnAyJ7oxdjq"
    did_key_p256_doc = {
        "@context": [
            "https://www.w3.org/ns/did/v1",
            "https://w3id.org/security/multikey/v1",
        ],
        "id": "did:key:zDnaehWviigWQQD7bqF3btFquwA5w8DX2sQwkVxnAyJ7oxdjq",
        "verificationMethod": [
            {
                "id": "did:key:zDnaehWviigWQQD7bqF3btFquwA5w8DX2sQwkVxnAyJ7oxdjq#zDnaehWviigWQQD7bqF3btFquwA5w8DX2sQwkVxnAyJ7oxdjq",
                "type": "Multikey",
                "controller": "did:key:zDnaehWviigWQQD7bqF3btFquwA5w8DX2sQwkVxnAyJ7oxdjq",
                "publicKeyMultibase": "zDnaehWviigWQQD7bqF3btFquwA5w8DX2sQwkVxnAyJ7oxdjq",
            }
        ],
        "authentication": [
            "did:key:zDnaehWviigWQQD7bqF3btFquwA5w8DX2sQwkVxnAyJ7oxdjq#zDnaehWviigWQQD7bqF3btFquwA5w8DX2sQwkVxnAyJ7oxdjq"
        ],
        "assertionMethod": [
            "did:key:zDnaehWviigWQQD7bqF3btFquwA5w8DX2sQwkVxnAyJ7oxdjq#zDnaehWviigWQQD7bqF3btFquwA5w8DX2sQwkVxnAyJ7oxdjq"
        ],
        "capabilityDelegation": [
            "did:key:zDnaehWviigWQQD7bqF3btFquwA5w8DX2sQwkVxnAyJ7oxdjq#zDnaehWviigWQQD7bqF3btFquwA5w8DX2sQwkVxnAyJ7oxdjq"
        ],
        "capabilityInvocation": [
            "did:key:zDnaehWviigWQQD7bqF3btFquwA5w8DX2sQwkVxnAyJ7oxdjq#zDnaehWviigWQQD7bqF3btFquwA5w8DX2sQwkVxnAyJ7oxdjq"
        ],
        "keyAgreement": [],
    }

    async def asyncSetUp(self):
        self.profile = await create_test_profile()
        self.profile.context.injector.bind_instance(DIDMethods, DIDMethods())
        self.profile.context.injector.bind_instance(
            BaseVerificationKeyStrategy, DefaultVerificationKeyStrategy()
        )
        self.profile.context.injector.bind_instance(KeyTypes, KeyTypes())

    async def setUpTestingDid(self, key_type: KeyType) -> Tuple[str, str]:
        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)
            await wallet.create_local_did(KEY, key_type, self.seed)

        if key_type == P256:
            did = self.did_key_p256_did
            vm_id = self.did_key_p256_verification_method
            did_doc = self.did_key_p256_doc
        elif key_type == ED25519:
            did = self.did_key_ed25519_did
            vm_id = self.did_key_ed25519_verification_method
            did_doc = self.did_key_ed25519_doc

        resolver = DIDResolver()
        resolver.register_resolver(
            MockResolver(
                ["key"],
                resolved=did_doc,
                native=True,
            )
        )
        self.profile.context.injector.bind_instance(DIDResolver, resolver)

        return (did, vm_id)

    async def test_sign_with_did_key_and_verify(self):
        for key_type in [ED25519, P256]:
            (did, _) = await self.setUpTestingDid(key_type)
            verification_method = None

            headers = {}
            payload = {}
            signed = await jwt_sign(
                self.profile, headers, payload, did, verification_method
            )

            assert signed

            assert await jwt_verify(self.profile, signed)

    async def test_sign_with_verification_method_and_verify(self):
        for key_type in [ED25519, P256]:
            (_, verification_method) = await self.setUpTestingDid(key_type)
            did = None
            headers = {}
            payload = {}
            signed: str = await jwt_sign(
                self.profile, headers, payload, did, verification_method
            )

            assert signed

            assert await jwt_verify(self.profile, signed)

    async def test_sign_x_invalid_did(self):
        did = "did:key:zzzzgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"
        headers = {}
        payload = {}
        verification_method = None
        with pytest.raises(Exception) as e_info:
            await jwt_sign(self.profile, headers, payload, did, verification_method)
        assert "No key type for prefixed public key" in str(e_info)

    async def test_sign_x_invalid_verification_method(self):
        did = None
        headers = {}
        payload = {}
        verification_method = "did:key:zzzzgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"
        with pytest.raises(Exception) as e_info:
            await jwt_sign(self.profile, headers, payload, did, verification_method)
        assert "Unknown DID" in str(e_info)

    async def test_verify_x_invalid_signed(self):
        for key_type in [ED25519, P256]:
            (did, _) = await self.setUpTestingDid(key_type)
            verification_method = None

            headers = {}
            payload = {}
            signed = await jwt_sign(
                self.profile, headers, payload, did, verification_method
            )

            assert signed
            signed = f"{signed[:-2]}2"

            with pytest.raises(Exception):
                await jwt_verify(self.profile, signed)

    async def test_resolve_public_key_by_kid_for_verify_ed25519(self):
        (_, kid) = await self.setUpTestingDid(ED25519)
        (key_bs58, key_type) = await resolve_public_key_by_kid_for_verify(
            self.profile, kid
        )

        assert key_bs58 == "3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRx"
        assert key_type == ED25519

    async def test_resolve_public_key_by_kid_for_verify_p256(self):
        (_, kid) = await self.setUpTestingDid(P256)
        (key_bs58, key_type) = await resolve_public_key_by_kid_for_verify(
            self.profile, kid
        )

        assert key_bs58 == "tYbR5egjfja9D5ix1jjYGqfh5QPu73RcZ7UjQUXtargj"
        assert key_type == P256



"""BbsBlsSignatureProof2020 class."""

import re
from os import urandom
from typing import List, Optional

from pyld import jsonld

from .bbs_bls_signature_2020_base import BbsBlsSignature2020Base

if BbsBlsSignature2020Base.BBS_SUPPORTED:
    from ursa_bbs_signatures import (
        BlsKeyPair,
        CreateProofRequest,
        ProofMessage,
        ProofMessageType,
        VerifyProofRequest,
        get_total_message_count,
    )
    from ursa_bbs_signatures import (
        create_proof as bls_create_proof,
    )
    from ursa_bbs_signatures import (
        verify_proof as bls_verify_proof,
    )

from ....utils.dependencies import assert_ursa_bbs_signatures_installed
from ....wallet.util import b64_to_bytes, bytes_to_b64
from ..crypto import _KeyPair as KeyPair
from ..document_loader import DocumentLoaderMethod
from ..error import LinkedDataProofException
from ..purposes import _ProofPurpose as ProofPurpose
from ..validation_result import ProofResult
from .bbs_bls_signature_2020 import BbsBlsSignature2020
from .linked_data_proof import DeriveProofResult


class BbsBlsSignatureProof2020(BbsBlsSignature2020Base):
    """BbsBlsSignatureProof2020 class."""

    signature_type = "BbsBlsSignatureProof2020"

    def __init__(
        self,
        *,
        key_pair: KeyPair,
    ):
        """Create new BbsBlsSignatureProof2020 instance.

        Args:
            key_pair (KeyPair): Key pair to use. Must provide BBS signatures

        """
        super().__init__(
            supported_derive_proof_types=(
                BbsBlsSignatureProof2020.supported_derive_proof_types
            ),
        )
        self.key_pair = key_pair
        self.mapped_derived_proof_type = "BbsBlsSignature2020"

    async def derive_proof(
        self,
        *,
        proof: dict,
        document: dict,
        reveal_document: dict,
        document_loader: DocumentLoaderMethod,
        nonce: Optional[bytes] = None,
    ):
        """Derive proof for document, return dict with derived document and proof."""
        assert_ursa_bbs_signatures_installed()

        # Validate that the input proof document has a proof compatible with this suite
        if proof.get("type") not in self.supported_derive_proof_types:
            raise LinkedDataProofException(
                f"Proof document proof incompatible, expected proof types of"
                f" {self.supported_derive_proof_types}, received " + proof["type"]
            )

        # Extract the BBS signature from the input proof
        signature = b64_to_bytes(proof["proofValue"])

        # Initialize the BBS signature suite
        # This is used for creating the input document verification data
        # NOTE: both suite._create_verify_xxx_data and self._create_verify_xxx_data
        # are used in this file. They have small changes in behavior
        suite = BbsBlsSignature2020(key_pair=self.key_pair)

        # Initialize the derived proof
        derived_proof = self.proof.copy() if self.proof else {}

        # Ensure proof type is set
        derived_proof["type"] = self.signature_type

        # Get the input document and proof statements
        document_statements = suite._create_verify_document_data(
            document=document, document_loader=document_loader
        )
        proof_statements = suite._create_verify_proof_data(
            proof=proof, document=document, document_loader=document_loader
        )

        # Transform any blank node identifiers for the input
        # document statements into actual node identifiers
        # e.g _:c14n0 => urn:bnid:_:c14n0
        transformed_input_document_statements = (
            self._transform_blank_node_ids_into_placeholder_node_ids(document_statements)
        )

        # Transform the resulting RDF statements back into JSON-LD
        compact_input_proof_document = jsonld.from_rdf(
            "\n".join(transformed_input_document_statements)
        )

        # Frame the result to create the reveal document result
        reveal_document_result = jsonld.frame(
            compact_input_proof_document,
            reveal_document,
            {"documentLoader": document_loader},
        )

        # Canonicalize the resulting reveal document
        reveal_document_statements = suite._create_verify_document_data(
            document=reveal_document_result, document_loader=document_loader
        )

        # Get the indices of the revealed statements from the transformed input document
        # offset by the number of proof statements
        number_of_proof_statements = len(proof_statements)

        # Always reveal all the statements associated to the original proof
        # these are always the first statements in the normalized form
        proof_reveal_indices = list(range(number_of_proof_statements))

        # Reveal the statements indicated from the reveal document
        document_reveal_indices = [
            transformed_input_document_statements.index(reveal_statement)
            + number_of_proof_statements
            for reveal_statement in reveal_document_statements
        ]

        # Check there is not a mismatch
        if len(document_reveal_indices) != len(reveal_document_statements):
            raise LinkedDataProofException(
                "Some statements in the reveal document not found in original proof"
            )

        # Combine all indices to get the resulting list of revealed indices
        reveal_indices = [*proof_reveal_indices, *document_reveal_indices]

        # Create a nonce if one is not supplied
        nonce = nonce or urandom(50)

        derived_proof["nonce"] = bytes_to_b64(
            nonce, urlsafe=False, pad=True, encoding="utf-8"
        )

        # Combine all the input statements that were originally signed
        # NOTE: we use plain strings here as input for the bbs lib.
        # the MATTR lib uses bytes, but the wrapper expects strings
        # it also works if we pass bytes as input
        all_input_statements = [*proof_statements, *document_statements]

        # Fetch the verification method
        verification_method = self._get_verification_method(
            proof=proof, document_loader=document_loader
        )

        # Create key pair from public key in verification method
        key_pair = self.key_pair.from_verification_method(verification_method)

        # Get the proof messages (revealed or not)
        proof_messages = []
        for input_statement_index in range(len(all_input_statements)):
            # if input statement index in revealed messages indexes use revealed type
            # otherwise use blinding
            proof_type = (
                ProofMessageType.Revealed
                if input_statement_index in reveal_indices
                else ProofMessageType.HiddenProofSpecificBlinding
            )
            proof_messages.append(
                ProofMessage(
                    message=all_input_statements[input_statement_index],
                    proof_type=proof_type,
                )
            )

        # get bbs key from bls key pair
        bbs_public_key = BlsKeyPair(public_key=key_pair.public_key).get_bbs_key(
            len(all_input_statements)
        )

        # Compute the proof
        proof_request = CreateProofRequest(
            public_key=bbs_public_key,
            messages=proof_messages,
            signature=signature,
            nonce=nonce,
        )

        output_proof = bls_create_proof(proof_request)

        # Set the proof value on the derived proof
        derived_proof["proofValue"] = bytes_to_b64(
            output_proof, urlsafe=False, pad=True, encoding="utf-8"
        )

        # Set the relevant proof elements on the derived proof from the input proof
        derived_proof["verificationMethod"] = proof["verificationMethod"]
        derived_proof["proofPurpose"] = proof["proofPurpose"]
        derived_proof["created"] = proof["created"]

        return DeriveProofResult(document={**reveal_document_result}, proof=derived_proof)

    async def verify_proof(
        self,
        *,
        proof: dict,
        document: dict,
        purpose: ProofPurpose,
        document_loader: DocumentLoaderMethod,
    ) -> ProofResult:
        """Verify proof against document and proof purpose."""
        assert_ursa_bbs_signatures_installed()
        try:
            proof["type"] = self.mapped_derived_proof_type

            # Get the proof and document statements
            proof_statements = self._create_verify_proof_data(
                proof=proof, document=document, document_loader=document_loader
            )
            document_statements = self._create_verify_document_data(
                document=document, document_loader=document_loader
            )

            # Transform the blank node identifier placeholders for the document statements
            # back into actual blank node identifiers
            transformed_document_statements = (
                self._transform_placeholder_node_ids_into_blank_node_ids(
                    document_statements
                )
            )

            # Combine all the statements to be verified
            # NOTE: we use plain strings here as input for the bbs lib.
            # the MATTR lib uses bytes, but the wrapper expects strings
            # it also works if we pass bytes as input
            statements_to_verify = [*proof_statements, *transformed_document_statements]

            # Fetch the verification method
            verification_method = self._get_verification_method(
                proof=proof, document_loader=document_loader
            )

            key_pair = self.key_pair.from_verification_method(verification_method)
            proof_bytes = b64_to_bytes(proof["proofValue"])

            total_message_count = get_total_message_count(proof_bytes)

            # get bbs key from bls key pair
            bbs_public_key = BlsKeyPair(public_key=key_pair.public_key).get_bbs_key(
                total_message_count
            )

            # verify derived proof
            verify_request = VerifyProofRequest(
                public_key=bbs_public_key,
                proof=proof_bytes,
                messages=statements_to_verify,
                nonce=b64_to_bytes(proof["nonce"]),
            )

            verified = bls_verify_proof(verify_request)

            if not verified:
                raise LinkedDataProofException(
                    f"Invalid signature on document {document}"
                )

            purpose_result = purpose.validate(
                proof=proof,
                document=document,
                suite=self,
                verification_method=verification_method,
                document_loader=document_loader,
            )

            if not purpose_result.valid:
                return ProofResult(
                    verified=False,
                    purpose_result=purpose_result,
                    error=purpose_result.error,
                )

            return ProofResult(verified=True, purpose_result=purpose_result)
        except Exception as err:
            return ProofResult(verified=False, error=err)

    def _canonize_proof(
        self, *, proof: dict, document: dict, document_loader: DocumentLoaderMethod
    ):
        """Canonize proof dictionary. Removes proofValue."""
        proof = {**proof, "@context": document.get("@context") or self._default_proof}

        proof.pop("proofValue", None)
        proof.pop("nonce", None)

        return self._canonize(input=proof, document_loader=document_loader)

    def _transform_blank_node_ids_into_placeholder_node_ids(
        self,
        statements: List[str],
    ) -> List[str]:
        """Transform blank node identifiers for the input into actual node identifiers.

        e.g _:c14n0 => urn:bnid:_:c14n0

        Args:
            statements (List[str]): List with possible blank node identifiers

        Returns:
            List[str]: List of transformed output statements

        """
        # replace all occurrences of _:c14nX with <urn:bnid:_:c14nX>
        transformed_statements = [
            re.sub(r"(_:c14n[0-9]+)", r"<urn:bnid:\1>", statement)
            for statement in statements
        ]

        return transformed_statements

    def _transform_placeholder_node_ids_into_blank_node_ids(
        self,
        statements: List[str],
    ) -> List[str]:
        """Transform the blank node placeholder identifiers back into actual blank nodes.

        e.g urn:bnid:_:c14n0 => _:c14n0

        Args:
            statements (List[str]): List with possible placeholder node identifiers

        Returns:
            List[str]: List of transformed output statements

        """
        # replace all occurrences of <urn:bnid:_:c14nX> with _:c14nX
        transformed_statements = [
            re.sub(r"<urn:bnid:(_:c14n[0-9]+)>", r"\1", statement)
            for statement in statements
        ]

        return transformed_statements

    supported_derive_proof_types = [
        "BbsBlsSignature2020",
        "sec:BbsBlsSignature2020",
        "https://w3id.org/security#BbsBlsSignature2020",
    ]



TEST_LD_DOCUMENT_SIGNED_BBS = {
    "@context": ["http://schema.org/", "https://w3id.org/security/bbs/v1"],
    "@type": "Person",
    "firstName": "Jane",
    "lastName": "Does",
    "jobTitle": "Professor",
    "telephone": "(425) 123-4567",
    "email": "jane.doe@example.com",
    "proof": {
        "type": "BbsBlsSignature2020",
        "created": "2020-10-30T04:34:10Z",
        "proofPurpose": "assertionMethod",
        "proofValue": "t+lSsiNnCYrVUzVsT1DoWMfZP5vHcXnrwo8kxSdsROv9+kTUjZWi5yEju4PZrHWwVPDP0AY2VYjDE5Y9QKYkxJ7PPg3CId2e/Wrk17Tjp31hEll7tBeMszRkTEUUb+gHSulYnYdN/Nvp/BWCekbuIg==",
        "verificationMethod": "did:example:489398593#test",
    },
}



import pytest
import pytest_asyncio

from .....core.event_bus import EventBus, MockEventBus
from .....messaging.request_context import RequestContext
from .....messaging.responder import MockResponder
from .....transport.inbound.receipt import MessageReceipt
from .....utils.testing import create_test_profile
from ..handler import ProblemReportHandler
from ..message import ProblemReport


@pytest_asyncio.fixture
async def request_context():
    yield RequestContext.test_context(await create_test_profile())


class TestPingHandler:
    @pytest.mark.asyncio
    async def test_problem_report(self, request_context):
        mock_event_bus = MockEventBus()
        request_context.profile.context.injector.bind_instance(EventBus, mock_event_bus)

        request_context.message_receipt = MessageReceipt()
        request_context.message = ProblemReport(description={"code": "error-code"})
        request_context.connection_ready = True
        handler = ProblemReportHandler()
        responder = MockResponder()
        await handler.handle(request_context, responder)
        messages = responder.messages
        assert len(messages) == 0
        assert len(mock_event_bus.events) == 1
        (profile, event) = mock_event_bus.events[0]
        assert profile == request_context.profile
        assert event.topic == "acapy::problem_report"
        assert event.payload == request_context.message.serialize()



"""Tests for Domain Txn Handling Utils."""

import json
from copy import deepcopy
from unittest import TestCase

import base58

from ..domain_txn_handler import (
    _extract_attr_typed_value,
    decode_state_value,
    extract_params_write_request,
    hash_of,
    make_state_path_for_attr,
    parse_attr_txn,
    prepare_attr_for_state,
    prepare_claim_def_for_state,
    prepare_get_claim_def_for_state,
    prepare_get_revoc_reg_entry_for_state,
    prepare_nym_for_state,
    prepare_revoc_def_for_state,
    prepare_revoc_reg_entry_accum_for_state,
    prepare_revoc_reg_entry_for_state,
    prepare_schema_for_state,
)

CLAIM_DEF_TXN = {
    "result": {
        "txn": {
            "data": {
                "ver": 1,
                "signature_type": "CL",
                "ref": 10,
                "tag": "some_tag",
                "data": {"primary": "....", "revocation": "...."},
            },
            "metadata": {
                "reqId": 1514280215504647,
                "from": "L5AD5g65TDQr1PPHHRoiGf",
                "endorser": "D6HG5g65TDQr1PPHHRoiGf",
                "digest": "6cee82226c6e276c983f46d03e3b3d10436d90b67bf33dc67ce9901b44dbc97c",
                "payloadDigest": "21f0f5c158ed6ad49ff855baf09a2ef9b4ed1a8015ac24bccc2e0106cd905685",
            },
        },
        "txnMetadata": {
            "txnTime": 1513945121,
            "seqNo": 10,
            "txnId": "HHAD5g65TDQr1PPHHRoiGf2L5AD5g65TDQr1PPHHRoiGf1|Degree1|CL|key1",
        },
    }
}

REVOC_REG_ENTRY_TXN = {
    "result": {
        "txn": {
            "data": {
                "ver": 1,
                "revocRegDefId": "L5AD5g65TDQr1PPHHRoiGf:3:FC4aWomrA13YyvYC1Mxw7:3:CL:14:some_tag:CL_ACCUM:tag1",
                "revocDefType": "CL_ACCUM",
                "value": {
                    "accum": "accum_value",
                    "prevAccum": "prev_acuum_value",
                    "issued": [],
                    "revoked": [10, 36, 3478],
                },
            },
        },
        "txnMetadata": {
            "txnTime": 1513945121,
            "seqNo": 10,
            "txnId": "5:L5AD5g65TDQr1PPHHRoiGf:3:FC4aWomrA13YyvYC1Mxw7:3:CL:14:some_tag:CL_ACCUM:tag1",
        },
    }
}


class TestDomainTxnHandler(TestCase):
    """Domain Txn Handler Tests"""

    def test_extract_attr_typed_value(self):
        test_txn_data = {"test": {...}}
        with self.assertRaises(ValueError) as cm:
            _extract_attr_typed_value(test_txn_data)
            assert "ATTR should have one" in cm
        test_txn_data = {
            "raw": {...},
            "enc": {...},
            "hash": {...},
        }
        with self.assertRaises(ValueError) as cm:
            _extract_attr_typed_value(test_txn_data)
            assert "ATTR should have only one" in cm

    def test_parse_attr_txn(self):
        test_txn_data = {"raw": '{"name": "Alice"}'}
        assert parse_attr_txn(test_txn_data) == ("raw", "name", '{"name": "Alice"}')
        test_txn_data = {"enc": "test"}
        assert parse_attr_txn(test_txn_data) == ("enc", "test", "test")
        test_txn_data = {"hash": "test"}
        assert parse_attr_txn(test_txn_data) == ("hash", "test", None)

    def test_decode_state_value(self):
        test_value = "test_value"
        test_lsn = "100"
        test_lut = "test_lut"
        test_encoded = {"val": test_value, "lsn": test_lsn, "lut": test_lut}
        assert decode_state_value(json.dumps(test_encoded)) == (
            test_value,
            test_lsn,
            test_lut,
        )

    def test_hash_of(self):
        test = {"test": "test"}
        assert hash_of(test)
        test = "123"
        assert hash_of(test)
        test = b"234"
        assert hash_of(test)

    def test_make_state_path_for_attr(self):
        assert b"did1:1:attrName1" == make_state_path_for_attr(
            "did1", "attrName1", attr_is_hash=True
        )
        assert (
            b"did1:1:677a81e8649df8f1a1e8af7709a5ece1d965cb684b2c185272114c5cc3b7ec49"
            == make_state_path_for_attr("did1", "attrName1", attr_is_hash=False)
        )
        assert (
            b"did1:1:677a81e8649df8f1a1e8af7709a5ece1d965cb684b2c185272114c5cc3b7ec49"
            == make_state_path_for_attr("did1", "attrName1")
        )

    def test_prepare_attr_for_state(self):
        txn = {
            "result": {
                "txn": {
                    "data": {
                        "ver": 1,
                        "dest": "N22KY2Dyvmuu2PyyqSFKue",
                        "raw": '{"name":"Alice"}',
                    },
                },
                "txnMetadata": {
                    "txnTime": 1513945121,
                    "seqNo": 10,
                    "txnId": "N22KY2Dyvmuu2PyyqSFKue|02",
                },
            }
        }
        txn = txn.get("result")
        path, value_bytes = prepare_attr_for_state(txn)
        assert (
            path
            == b"N22KY2Dyvmuu2PyyqSFKue:1:82a3537ff0dbce7eec35d69edc3a189ee6f17d82f353a553f9aa96cb0be3ce89"
        )
        assert (
            value_bytes
            == b'{"lsn": 10, "lut": 1513945121, "val": "6d4a333838d0ef96756cccc680af2531075c512502fb68c5503c63d93de859b3"}'
        )
        path = prepare_attr_for_state(txn, path_only=True)
        assert (
            path
            == b"N22KY2Dyvmuu2PyyqSFKue:1:82a3537ff0dbce7eec35d69edc3a189ee6f17d82f353a553f9aa96cb0be3ce89"
        )

    def test_prepare_nym_for_state(self):
        txn = {
            "result": {
                "txn": {
                    "data": {
                        "dest": "N22KY2Dyvmuu2PyyqSFKue",
                    },
                },
            }
        }
        txn = txn.get("result")
        assert prepare_nym_for_state(txn)

    def test_prepare_schema_for_state(self):
        txn = {
            "result": {
                "txn": {
                    "type": "101",
                    "protocolVersion": 2,
                    "data": {
                        "ver": 1,
                        "data": {
                            "name": "Degree",
                            "version": "1.0",
                            "attr_names": [
                                "undergrad",
                                "last_name",
                                "first_name",
                                "birth_date",
                                "postgrad",
                                "expiry_date",
                            ],
                        },
                    },
                    "metadata": {
                        "reqId": 1514280215504647,
                        "from": "L5AD5g65TDQr1PPHHRoiGf",
                        "endorser": "D6HG5g65TDQr1PPHHRoiGf",
                        "digest": "6cee82226c6e276c983f46d03e3b3d10436d90b67bf33dc67ce9901b44dbc97c",
                        "payloadDigest": "21f0f5c158ed6ad49ff855baf09a2ef9b4ed1a8015ac24bccc2e0106cd905685",
                    },
                },
                "txnMetadata": {
                    "txnTime": 1513945121,
                    "seqNo": 10,
                    "txnId": "L5AD5g65TDQr1PPHHRoiGf1|Degree|1.0",
                },
            }
        }
        txn = txn.get("result")
        path, value_bytes = prepare_schema_for_state(txn)
        assert path == b"L5AD5g65TDQr1PPHHRoiGf:2:Degree:1.0"
        assert (
            prepare_schema_for_state(txn, path_only=True)
            == b"L5AD5g65TDQr1PPHHRoiGf:2:Degree:1.0"
        )

    def test_prepare_get_claim_def_for_state(self):
        txn = deepcopy(CLAIM_DEF_TXN)
        txn.get("result").get("txn").get("data").pop("ref")
        with self.assertRaises(ValueError) as cm:
            prepare_get_claim_def_for_state(txn)
            assert "ref field is absent, but it must contain schema seq no" in cm

    def test_prepare_claim_def_for_state(self):
        txn = deepcopy(CLAIM_DEF_TXN.get("result"))
        txn.get("txn").get("data").pop("ref")
        with self.assertRaises(ValueError) as cm:
            prepare_claim_def_for_state(txn)
            assert "ref field is absent, but it must contain schema seq no" in cm

        txn = deepcopy(CLAIM_DEF_TXN.get("result"))
        txn.get("txn").get("data").pop("data")
        with self.assertRaises(ValueError) as cm:
            prepare_claim_def_for_state(txn)
            assert "data field is absent, but it must contain components of keys" in cm

        txn = deepcopy(CLAIM_DEF_TXN.get("result"))
        path, value_bytes = prepare_claim_def_for_state(txn)
        assert path == b"L5AD5g65TDQr1PPHHRoiGf:3:CL:10:some_tag"
        assert (
            prepare_claim_def_for_state(txn, path_only=True)
            == b"L5AD5g65TDQr1PPHHRoiGf:3:CL:10:some_tag"
        )

    def test_prepare_revoc_def_for_state(self):
        txn = {
            "result": {
                "txn": {
                    "type": "113",
                    "protocolVersion": 2,
                    "data": {
                        "ver": 1,
                        "id": "L5AD5g65TDQr1PPHHRoiGf:3:FC4aWomrA13YyvYC1Mxw7:3:CL:14:some_tag:CL_ACCUM:tag1",
                        "credDefId": "FC4aWomrA13YyvYC1Mxw7:3:CL:14:some_tag",
                        "revocDefType": "CL_ACCUM",
                        "tag": "tag1",
                        "value": {
                            "maxCredNum": 1000000,
                            "tailsHash": "6619ad3cf7e02fc29931a5cdc7bb70ba4b9283bda3badae297",
                            "tailsLocation": "http://tails.location.com",
                            "issuanceType": "ISSUANCE_BY_DEFAULT",
                            "publicKeys": {},
                        },
                    },
                    "metadata": {
                        "reqId": 1514280215504647,
                        "from": "L5AD5g65TDQr1PPHHRoiGf",
                        "endorser": "D6HG5g65TDQr1PPHHRoiGf",
                        "digest": "6cee82226c6e276c983f46d03e3b3d10436d90b67bf33dc67ce9901b44dbc97c",
                        "payloadDigest": "21f0f5c158ed6ad49ff855baf09a2ef9b4ed1a8015ac24bccc2e0106cd905685",
                    },
                },
                "txnMetadata": {
                    "txnTime": 1513945121,
                    "seqNo": 10,
                    "txnId": "L5AD5g65TDQr1PPHHRoiGf:3:FC4aWomrA13YyvYC1Mxw7:3:CL:14:some_tag:CL_ACCUM:tag1",
                },
            },
        }
        txn = txn.get("result")
        path, value_bytes = prepare_revoc_def_for_state(txn)
        assert (
            path
            == b"L5AD5g65TDQr1PPHHRoiGf:4:FC4aWomrA13YyvYC1Mxw7:3:CL:14:some_tag:CL_ACCUM:tag1"
        )
        assert (
            prepare_revoc_def_for_state(txn, path_only=True)
            == b"L5AD5g65TDQr1PPHHRoiGf:4:FC4aWomrA13YyvYC1Mxw7:3:CL:14:some_tag:CL_ACCUM:tag1"
        )

    def test_prepare_get_revoc_reg_entry_for_state(self):
        assert prepare_get_revoc_reg_entry_for_state(REVOC_REG_ENTRY_TXN)
        assert prepare_get_revoc_reg_entry_for_state(REVOC_REG_ENTRY_TXN.get("result"))

    def test_prepare_revoc_reg_entry_for_state(self):
        path, value_bytes = prepare_revoc_reg_entry_for_state(
            REVOC_REG_ENTRY_TXN.get("result")
        )
        assert (
            path
            == b"5:L5AD5g65TDQr1PPHHRoiGf:3:FC4aWomrA13YyvYC1Mxw7:3:CL:14:some_tag:CL_ACCUM:tag1"
        )
        assert (
            prepare_revoc_reg_entry_for_state(
                REVOC_REG_ENTRY_TXN.get("result"), path_only=True
            )
            == b"5:L5AD5g65TDQr1PPHHRoiGf:3:FC4aWomrA13YyvYC1Mxw7:3:CL:14:some_tag:CL_ACCUM:tag1"
        )

    def test_prepare_revoc_reg_entry_accum_for_state(self):
        path, value_bytes = prepare_revoc_reg_entry_accum_for_state(REVOC_REG_ENTRY_TXN)
        assert (
            path
            == b"6:L5AD5g65TDQr1PPHHRoiGf:3:FC4aWomrA13YyvYC1Mxw7:3:CL:14:some_tag:CL_ACCUM:tag1"
        )
        path, value_bytes = prepare_revoc_reg_entry_accum_for_state(
            REVOC_REG_ENTRY_TXN.get("result")
        )
        assert (
            path
            == b"6:L5AD5g65TDQr1PPHHRoiGf:3:FC4aWomrA13YyvYC1Mxw7:3:CL:14:some_tag:CL_ACCUM:tag1"
        )

    def test_extract_params_write_request(self):
        write_request = {
            "result": {
                "txn": {
                    "data": {
                        "data": {
                            "attr_names": ["axuall_proof_id"],
                            "name": "manual_issuance",
                            "version": "4.1.0",
                        }
                    },
                    "metadata": {
                        "digest": "f440da62ab1c38601e41b5e148370b64af3183a2428f327cfb8ac83ca4cbc698",
                        "from": "F72i3Y3Q4i466efjYJYCHM",
                        "payloadDigest": "0ecfd475ed34635b6623bcbcda422ff90dee53f66de20a7441bca04ba0632670",
                        "reqId": 1634854863739554300,
                        "taaAcceptance": {
                            "mechanism": "for_session",
                            "taaDigest": "8cee5d7a573e4893b08ff53a0761a22a1607df3b3fcd7e75b98696c92879641f",
                            "time": 1600300800,
                        },
                    },
                    "protocolVersion": 2,
                    "type": "101",
                },
                "txnMetadata": {
                    "seqNo": 75335,
                    "txnId": "F72i3Y3Q4i466efjYJYCHM:2:manual_issuance:4.1.0",
                    "txnTime": 1634854863,
                },
                "ver": "1",
                "auditPath": [
                    "DhzuS5ZkyxUrQbRzbCfq3NDcTnTzFcJHYJYyTAqgMymT",
                    "AtYz5H13m1ZyFiGVVLwRL7au8muhCMwikeQrMwJBJ5d9",
                    "3GpqpZw4jymXDFm8Mf5gSMHSJ8r1wXutiZ1x26kxwWBE",
                    "EQyUpv9z64prPmDqkb8n2SzBtor3tGaNw4qZP2psG2wD",
                    "9fbpJjWbDuNbCrC7GQxjVxDGJ48k7rdBDf5ATu7PXVx3",
                    "CRbATGbi9DbN1pphXMcREVZTUUqsQbD9mN5ikKm1FkSc",
                    "E19R3Cty6iLtWBoxSqHLrxTSJSKpBtj3wJEdD98pS9g8",
                ],
                "ledgerSize": 75335,
                "reqSignature": {
                    "type": "ED25519",
                    "values": [
                        {
                            "from": "F72i3Y3Q4i466efjYJYCHM",
                            "value": "29sKMwQjm6r2BjpbJdFi1TGvowaxbi9nnAPP9LwPHhuBzwtfWwqJFg4Ur2xAWBLYfPLnMNNSgwwNesRSb1C1L72d",
                        }
                    ],
                },
                "rootHash": "4aM6yCpamk82Uqb414mNRpEdYkwdSMhi3HkXgN7YvRaX",
            }
        }
        txn = deepcopy(write_request)
        (
            tree_size,
            leaf_index,
            decoded_audit_path,
            expected_root_hash,
        ) = extract_params_write_request(txn)

        given_audit_path = [
            "DhzuS5ZkyxUrQbRzbCfq3NDcTnTzFcJHYJYyTAqgMymT",
            "AtYz5H13m1ZyFiGVVLwRL7au8muhCMwikeQrMwJBJ5d9",
            "3GpqpZw4jymXDFm8Mf5gSMHSJ8r1wXutiZ1x26kxwWBE",
            "EQyUpv9z64prPmDqkb8n2SzBtor3tGaNw4qZP2psG2wD",
            "9fbpJjWbDuNbCrC7GQxjVxDGJ48k7rdBDf5ATu7PXVx3",
            "CRbATGbi9DbN1pphXMcREVZTUUqsQbD9mN5ikKm1FkSc",
            "E19R3Cty6iLtWBoxSqHLrxTSJSKpBtj3wJEdD98pS9g8",
        ]
        expected_audit_path = [
            base58.b58decode(hash_str.encode("utf-8")) for hash_str in given_audit_path
        ]
        expected_hash = base58.b58decode(
            "4aM6yCpamk82Uqb414mNRpEdYkwdSMhi3HkXgN7YvRaX".encode("utf-8")
        )
        assert tree_size == 75335
        assert leaf_index == 75334
        assert expected_root_hash == expected_hash
        assert decoded_audit_path == expected_audit_path

        txn = deepcopy(write_request)
        txn = txn.get("result")
        (
            tree_size,
            leaf_index,
            decoded_audit_path,
            expected_root_hash,
        ) = extract_params_write_request(txn)

        assert tree_size == 75335
        assert leaf_index == 75334
        assert expected_root_hash == expected_hash
        assert decoded_audit_path == expected_audit_path

        txn = deepcopy(write_request)
        txn["result"]["txnMetadata"]["seqNo"] = 1000
        with self.assertRaises(Exception) as cm:
            extract_params_write_request(txn)
            assert "auditPath length does not match with given seqNo" in cm



"""."""

import logging
from typing import Any, Mapping, Optional, Union

from marshmallow import fields

from .....core.profile import ProfileSession
from .....messaging.models.base_record import BaseExchangeRecord, BaseExchangeSchema
from .....messaging.valid import UUID4_EXAMPLE
from .....storage.error import StorageDuplicateError, StorageNotFoundError
from ..messages.disclose import Disclose, DiscloseSchema
from ..messages.query import Query, QuerySchema
from . import UNENCRYPTED_TAGS

LOGGER = logging.getLogger(__name__)


class V10DiscoveryExchangeRecord(BaseExchangeRecord):
    """Represents a Discover Feature (0031) exchange record."""

    class Meta:
        """V10DiscoveryExchangeRecord metadata."""

        schema_class = "V10DiscoveryRecordSchema"

    RECORD_TYPE = "discovery_exchange_v10"
    RECORD_ID_NAME = "discovery_exchange_id"
    RECORD_TOPIC = "discover_feature"
    TAG_NAMES = {"~thread_id" if UNENCRYPTED_TAGS else "thread_id", "connection_id"}

    STATE_QUERY_SENT = "query-sent"
    STATE_DISCLOSE_RECV = "disclose-received"

    def __init__(
        self,
        *,
        state: Optional[str] = None,
        discovery_exchange_id: Optional[str] = None,
        connection_id: Optional[str] = None,
        thread_id: Optional[str] = None,
        query_msg: Union[Mapping, Query, None] = None,
        disclose: Union[Mapping, Disclose, None] = None,
        **kwargs,
    ):
        """Initialize a new V10DiscoveryExchangeRecord."""
        super().__init__(discovery_exchange_id, state or self.STATE_QUERY_SENT, **kwargs)
        self._id = discovery_exchange_id
        self.connection_id = connection_id
        self.thread_id = thread_id
        self._query_msg = Query.serde(query_msg)
        self._disclose = Disclose.serde(disclose)

    @property
    def discovery_exchange_id(self) -> str:
        """Accessor for the ID."""
        return self._id

    @property
    def query_msg(self) -> Query:
        """Accessor; get deserialized view."""
        return None if self._query_msg is None else self._query_msg.de

    @query_msg.setter
    def query_msg(self, value):
        """Setter; store de/serialized views."""
        self._query_msg = Query.serde(value)

    @property
    def disclose(self) -> Disclose:
        """Accessor; get deserialized view."""
        return None if self._disclose is None else self._disclose.de

    @disclose.setter
    def disclose(self, value):
        """Setter; store de/serialized views."""
        self._disclose = Disclose.serde(value)

    @classmethod
    async def retrieve_by_connection_id(
        cls, session: ProfileSession, connection_id: str
    ) -> "V10DiscoveryExchangeRecord":
        """Retrieve a discovery exchange record by connection."""
        tag_filter = {"connection_id": connection_id}
        return await cls.retrieve_by_tag_filter(session, tag_filter)

    @classmethod
    async def retrieve_if_exists_by_connection_id(
        cls, session: ProfileSession, connection_id: str
    ) -> Optional["V10DiscoveryExchangeRecord"]:
        """Retrieve a discovery exchange record by connection."""
        tag_filter = {"connection_id": connection_id}
        result = await cls.query(session, tag_filter)
        if len(result) > 1:
            LOGGER.warning(
                "More than one disclosure record found for connection: %s",
                connection_id,
            )

        return result[0] if result else None

    @classmethod
    async def exists_for_connection_id(
        cls, session: ProfileSession, connection_id: str
    ) -> bool:
        """Return whether a discovery exchange record exists for the given connection.

        Args:
            session (ProfileSession): session
            connection_id (str): connection_id

        Returns:
            bool: whether record exists

        """
        tag_filter = {"connection_id": connection_id}
        try:
            record = await cls.retrieve_by_tag_filter(session, tag_filter)
        except StorageNotFoundError:
            return False
        except StorageDuplicateError:
            return True
        return bool(record)

    @property
    def record_value(self) -> dict:
        """Accessor for the JSON record value generated."""
        return {
            **{
                prop: getattr(self, f"_{prop}").ser
                for prop in (
                    "query_msg",
                    "disclose",
                )
                if getattr(self, prop) is not None
            },
        }

    def __eq__(self, other: Any) -> bool:
        """Comparison between records."""
        return super().__eq__(other)


class V10DiscoveryRecordSchema(BaseExchangeSchema):
    """Schema to allow ser/deser of Discover Feature (0031) records."""

    class Meta:
        """V10DiscoveryRecordSchema metadata."""

        model_class = V10DiscoveryExchangeRecord

    discovery_exchange_id = fields.Str(
        required=False,
        metadata={
            "description": "Credential exchange identifier",
            "example": UUID4_EXAMPLE,
        },
    )
    connection_id = fields.Str(
        required=False,
        metadata={"description": "Connection identifier", "example": UUID4_EXAMPLE},
    )
    thread_id = fields.Str(
        required=False,
        metadata={"description": "Thread identifier", "example": UUID4_EXAMPLE},
    )
    query_msg = fields.Nested(
        QuerySchema(), required=False, metadata={"description": "Query message"}
    )
    disclose = fields.Nested(
        DiscloseSchema(), required=False, metadata={"description": "Disclose message"}
    )



import json
from unittest import IsolatedAsyncioTestCase

import pytest
from aiohttp import web

from ...admin.request_context import AdminRequestContext
from ...anoncreds.base import AnonCredsObjectNotFound
from ...anoncreds.issuer import AnonCredsIssuer
from ...anoncreds.models.schema import AnonCredsSchema, SchemaResult, SchemaState
from ...anoncreds.revocation import AnonCredsRevocation
from ...anoncreds.revocation_setup import DefaultRevocationSetup
from ...core.event_bus import MockEventBus
from ...tests import mock
from ...utils.testing import create_test_profile
from .. import routes as test_module


class MockSchema:
    def __init__(self, schema_id):
        self.schemaId = schema_id

    def serialize(self):
        return {"schema_id": self.schemaId}


class MockCredentialDefinition:
    def __init__(self, cred_def_id):
        self.credDefId = cred_def_id

    def serialize(self):
        return {"credential_definition_id": self.credDefId}


class MockRovocationRegistryDefinition:
    def __init__(self, rev_reg_id):
        self.revRegId = rev_reg_id

    def serialize(self):
        return {"revocation_registry_definition_id": self.revRegId}


@pytest.mark.anoncreds
class TestAnonCredsRoutes(IsolatedAsyncioTestCase):
    async def asyncSetUp(self) -> None:
        self.session_inject = {}
        self.profile = await create_test_profile(
            settings={
                "wallet.type": "askar-anoncreds",
                "admin.admin_api_key": "secret-key",
            },
        )
        self.context = AdminRequestContext.test_context(self.session_inject, self.profile)
        self.request_dict = {
            "context": self.context,
        }
        self.request = mock.MagicMock(
            app={},
            match_info={},
            query={},
            __getitem__=lambda _, k: self.request_dict[k],
            context=self.context,
            headers={"x-api-key": "secret-key"},
        )

    @mock.patch.object(
        AnonCredsIssuer,
        "create_and_register_schema",
        return_value=SchemaResult(
            job_id=None,
            schema_state=SchemaState(
                state="finished",
                schema_id=None,
                schema=AnonCredsSchema(
                    issuer_id="issuer-id",
                    name="name",
                    version="1.0",
                    attr_names=["attr1", "attr2"],
                ),
            ),
        ),
    )
    async def test_schemas_post(self, mock_create_and_register_schema):
        self.request.json = mock.CoroutineMock(
            side_effect=[
                {
                    "schema": {
                        "issuerId": "Q4TmbeGPoWeWob4Xf6KetA",
                        "attrNames": ["score"],
                        "name": "Example Schema",
                        "version": "0.0.1",
                    }
                },
                {},
                {"schema": {}},
                {
                    "schema": {
                        "attrNames": ["score"],
                        "name": "Example Schema",
                        "version": "0.0.1",
                    }
                },
            ]
        )
        result = await test_module.schemas_post(self.request)
        assert result is not None

        assert mock_create_and_register_schema.call_count == 1

        with self.assertRaises(web.HTTPBadRequest):
            # Empty body
            await test_module.schemas_post(self.request)
            # Empty schema
            await test_module.schemas_post(self.request)
            # Missing issuerId
            await test_module.schemas_post(self.request)

    async def test_get_schema(self):
        self.request.match_info = {"schema_id": "schema_id"}
        self.context.inject = mock.Mock(
            return_value=mock.MagicMock(
                get_schema=mock.CoroutineMock(
                    side_effect=[
                        MockSchema("schemaId"),
                        AnonCredsObjectNotFound("test"),
                    ]
                )
            )
        )
        result = await test_module.schema_get(self.request)
        assert json.loads(result.body)["schema_id"] == "schemaId"

        # missing schema_id
        self.request.match_info = {}
        with self.assertRaises(KeyError):
            await test_module.schema_get(self.request)

        # schema not found
        self.request.match_info = {"schema_id": "schema_id"}
        with self.assertRaises(web.HTTPNotFound):
            await test_module.schema_get(self.request)

    @mock.patch.object(
        AnonCredsIssuer,
        "get_created_schemas",
        side_effect=[
            [
                "Q4TmbeGPoWeWob4Xf6KetA:2:Example Schema:0.0.1",
                "Q4TmbeGPoWeWob4Xf6KetA:2:Example Schema:0.0.2",
            ],
            [],
        ],
    )
    async def test_get_schemas(self, mock_get_created_schemas):
        result = await test_module.schemas_get(self.request)
        assert json.loads(result.body)["schema_ids"].__len__() == 2

        result = await test_module.schemas_get(self.request)
        assert json.loads(result.body)["schema_ids"].__len__() == 0

        assert mock_get_created_schemas.call_count == 2

    @mock.patch.object(
        AnonCredsIssuer,
        "create_and_register_credential_definition",
        return_value=MockCredentialDefinition("credDefId"),
    )
    async def test_cred_def_post(self, mock_create_cred_def):
        self.request.json = mock.CoroutineMock(
            side_effect=[
                {
                    "credential_definition": {
                        "issuerId": "issuerId",
                        "schemaId": "schemaId",
                        "tag": "tag",
                    },
                    "options": {
                        "endorser_connection_id": "string",
                        "revocation_registry_size": 0,
                        "support_revocation": True,
                    },
                },
                {},
                {"credential_definition": {}},
            ]
        )

        result = await test_module.cred_def_post(self.request)

        assert json.loads(result.body)["credential_definition_id"] == "credDefId"
        assert mock_create_cred_def.call_count == 1

        with self.assertRaises(web.HTTPBadRequest):
            await test_module.cred_def_post(self.request)

        await test_module.cred_def_post(self.request)

    async def test_cred_def_get(self):
        self.request.match_info = {"cred_def_id": "cred_def_id"}
        self.context.inject = mock.Mock(
            return_value=mock.MagicMock(
                get_credential_definition=mock.CoroutineMock(
                    return_value=MockCredentialDefinition("credDefId")
                )
            )
        )
        result = await test_module.cred_def_get(self.request)
        assert json.loads(result.body)["credential_definition_id"] == "credDefId"

        self.request.match_info = {}
        with self.assertRaises(KeyError):
            await test_module.cred_def_get(self.request)

    @mock.patch.object(
        AnonCredsIssuer,
        "get_created_credential_definitions",
        side_effect=[
            [
                "Q4TmbeGPoWeWob4Xf6KetA:3:CL:229927:tag",
                "Q4TmbeGPoWeWob4Xf6KetA:3:CL:229925:faber.agent.degree_schema",
            ],
            [],
        ],
    )
    async def test_cred_defs_get(self, mock_get_cred_defs):
        result = await test_module.cred_defs_get(self.request)
        assert len(json.loads(result.body)["credential_definition_ids"]) == 2

        result = await test_module.cred_defs_get(self.request)
        assert len(json.loads(result.body)["credential_definition_ids"]) == 0

        assert mock_get_cred_defs.call_count == 2

    @mock.patch.object(
        AnonCredsIssuer,
        "match_created_credential_definitions",
        side_effect=["found", None],
    )
    @mock.patch.object(
        AnonCredsRevocation,
        "create_and_register_revocation_registry_definition",
        return_value=MockRovocationRegistryDefinition("revRegId"),
    )
    async def test_rev_reg_def_post(self, mock_match, mock_create):
        self.request.json = mock.CoroutineMock(
            return_value={
                "credDefId": "cred_def_id",
                "issuerId": "issuer_id",
                "maxCredNum": 100,
                "options": {
                    "tails_public_uri": "http://tails_public_uri",
                    "tails_local_uri": "http://tails_local_uri",
                },
            }
        )

        # Must be in wrapper object
        with self.assertRaises(web.HTTPBadRequest):
            await test_module.rev_reg_def_post(self.request)

        self.request.json = mock.CoroutineMock(
            return_value={
                "revocation_registry_definition": {
                    "credDefId": "cred_def_id",
                    "issuerId": "issuer_id",
                    "maxCredNum": 100,
                    "options": {
                        "tails_public_uri": "http://tails_public_uri",
                        "tails_local_uri": "http://tails_local_uri",
                    },
                }
            }
        )

        result = await test_module.rev_reg_def_post(self.request)

        assert json.loads(result.body)["revocation_registry_definition_id"] == "revRegId"

        assert mock_match.call_count == 1
        assert mock_create.call_count == 1

        with self.assertRaises(web.HTTPNotFound):
            await test_module.rev_reg_def_post(self.request)

    @mock.patch.object(
        AnonCredsRevocation,
        "create_and_register_revocation_list",
        return_value=MockRovocationRegistryDefinition("revRegId"),
    )
    async def test_rev_list_post(self, mock_create):
        self.request.json = mock.CoroutineMock(
            return_value={"revRegDefId": "rev_reg_def_id", "options": {}}
        )
        result = await test_module.rev_list_post(self.request)
        assert json.loads(result.body)["revocation_registry_definition_id"] == "revRegId"
        assert mock_create.call_count == 1

    @mock.patch.object(
        AnonCredsRevocation,
        "get_created_revocation_registry_definition",
        side_effect=[
            MockRovocationRegistryDefinition("revRegId"),
            None,
            MockRovocationRegistryDefinition("revRegId"),
        ],
    )
    @mock.patch.object(
        AnonCredsRevocation,
        "upload_tails_file",
        return_value=None,
    )
    async def test_upload_tails_file(self, mock_upload, mock_get):
        self.request.match_info = {"rev_reg_id": "rev_reg_id"}
        result = await test_module.upload_tails_file(self.request)
        assert result is not None
        assert mock_upload.call_count == 1
        assert mock_get.call_count == 1

        with self.assertRaises(web.HTTPNotFound):
            await test_module.upload_tails_file(self.request)

        self.request.match_info = {}

        with self.assertRaises(KeyError):
            await test_module.upload_tails_file(self.request)

    @mock.patch.object(
        AnonCredsRevocation,
        "set_active_registry",
        return_value=None,
    )
    async def test_set_active_registry(self, mock_set):
        self.request.match_info = {"rev_reg_id": "rev_reg_id"}
        await test_module.set_active_registry(self.request)
        assert mock_set.call_count == 1

        self.request.match_info = {}
        with self.assertRaises(KeyError):
            await test_module.set_active_registry(self.request)

    async def test_schema_endpoints_wrong_profile_403(self):
        self.profile = await create_test_profile(
            settings={"wallet-type": "askar", "admin.admin_api_key": "secret-key"},
        )
        self.context = AdminRequestContext.test_context({}, self.profile)
        self.request_dict = {
            "context": self.context,
        }
        self.request = mock.MagicMock(
            app={},
            match_info={},
            query={},
            __getitem__=lambda _, k: self.request_dict[k],
            context=self.context,
            headers={"x-api-key": "secret-key"},
        )

        # POST schema
        self.request.json = mock.CoroutineMock(
            return_value={
                "schema": {
                    "issuerId": "Q4TmbeGPoWeWob4Xf6KetA",
                    "attrNames": ["score"],
                    "name": "Example Schema",
                    "version": "0.0.1",
                }
            }
        )
        with self.assertRaises(web.HTTPForbidden):
            await test_module.schemas_post(self.request)

        # GET schema
        self.request.match_info = {"schema_id": "schema_id"}
        with self.assertRaises(web.HTTPForbidden):
            await test_module.schema_get(self.request)

        # GET schemas
        with self.assertRaises(web.HTTPForbidden):
            await test_module.schemas_get(self.request)

    async def test_cred_def_endpoints_wrong_profile_403(self):
        self.profile = await create_test_profile(
            settings={"wallet-type": "askar", "admin.admin_api_key": "secret-key"},
        )
        self.context = AdminRequestContext.test_context({}, self.profile)
        self.request_dict = {
            "context": self.context,
        }
        self.request = mock.MagicMock(
            app={},
            match_info={},
            query={},
            __getitem__=lambda _, k: self.request_dict[k],
            context=self.context,
            headers={"x-api-key": "secret-key"},
        )

        # POST cred def
        self.request.json = mock.CoroutineMock(
            return_value={
                "credential_definition": {
                    "issuerId": "issuerId",
                    "schemaId": "schemaId",
                    "tag": "tag",
                },
                "options": {
                    "revocation_registry_size": 0,
                    "support_revocation": True,
                },
            }
        )
        with self.assertRaises(web.HTTPForbidden):
            await test_module.cred_def_post(self.request)

        # GET cred def
        self.request.match_info = {"cred_def_id": "cred_def_id"}
        with self.assertRaises(web.HTTPForbidden):
            await test_module.cred_def_get(self.request)

        # GET cred defs
        with self.assertRaises(web.HTTPForbidden):
            await test_module.cred_defs_get(self.request)

    async def test_rev_reg_wrong_profile_403(self):
        self.profile = await create_test_profile(
            settings={"wallet-type": "askar", "admin.admin_api_key": "secret-key"},
        )
        self.context = AdminRequestContext.test_context({}, self.profile)
        self.request_dict = {
            "context": self.context,
        }
        self.request = mock.MagicMock(
            app={},
            match_info={},
            query={},
            __getitem__=lambda _, k: self.request_dict[k],
            context=self.context,
            headers={"x-api-key": "secret-key"},
        )

        self.request.json = mock.CoroutineMock(
            return_value={
                "revocation_registry_definition": {
                    "credDefId": "cred_def_id",
                    "issuerId": "issuer_id",
                    "maxCredNum": 100,
                },
                "options": {
                    "tails_public_uri": "http://tails_public_uri",
                    "tails_local_uri": "http://tails_local_uri",
                },
            }
        )
        with self.assertRaises(web.HTTPForbidden):
            await test_module.rev_reg_def_post(self.request)

    async def test_rev_list_wrong_profile_403(self):
        self.profile = await create_test_profile(
            settings={"wallet-type": "askar", "admin.admin_api_key": "secret-key"},
        )
        self.context = AdminRequestContext.test_context({}, self.profile)
        self.request_dict = {
            "context": self.context,
        }
        self.request = mock.MagicMock(
            app={},
            match_info={},
            query={},
            __getitem__=lambda _, k: self.request_dict[k],
            context=self.context,
            headers={"x-api-key": "secret-key"},
        )

        self.request.json = mock.CoroutineMock(
            return_value={"revRegDefId": "rev_reg_def_id", "options": {}}
        )
        with self.assertRaises(web.HTTPForbidden):
            await test_module.rev_list_post(self.request)

    async def test_uploads_tails_wrong_profile_403(self):
        self.profile = await create_test_profile(
            settings={"wallet-type": "askar", "admin.admin_api_key": "secret-key"},
        )
        self.context = AdminRequestContext.test_context({}, self.profile)
        self.request_dict = {
            "context": self.context,
        }
        self.request = mock.MagicMock(
            app={},
            match_info={},
            query={},
            __getitem__=lambda _, k: self.request_dict[k],
            context=self.context,
            headers={"x-api-key": "secret-key"},
        )

        self.request.match_info = {"rev_reg_id": "rev_reg_id"}
        with self.assertRaises(web.HTTPForbidden):
            await test_module.upload_tails_file(self.request)

    async def test_active_registry_wrong_profile_403(self):
        self.profile = await create_test_profile(
            settings={"wallet-type": "askar", "admin.admin_api_key": "secret-key"},
        )
        self.context = AdminRequestContext.test_context({}, self.profile)
        self.request_dict = {
            "context": self.context,
        }
        self.request = mock.MagicMock(
            app={},
            match_info={},
            query={},
            __getitem__=lambda _, k: self.request_dict[k],
            context=self.context,
            headers={"x-api-key": "secret-key"},
        )

        self.request.match_info = {"rev_reg_id": "rev_reg_id"}

        with self.assertRaises(web.HTTPForbidden):
            await test_module.set_active_registry(self.request)

    @mock.patch.object(DefaultRevocationSetup, "register_events")
    async def test_register_events(self, mock_revocation_setup_listeners):
        mock_event_bus = MockEventBus()
        mock_event_bus.subscribe = mock.MagicMock()
        test_module.register_events(mock_event_bus)
        assert mock_revocation_setup_listeners.call_count == 1

    async def test_register(self):
        mock_app = mock.MagicMock()
        mock_app.add_routes = mock.MagicMock()

        await test_module.register(mock_app)
        mock_app.add_routes.assert_called_once()

    async def test_post_process_routes(self):
        mock_app = mock.MagicMock(_state={"swagger_dict": {}})
        test_module.post_process_routes(mock_app)
        assert "tags" in mock_app._state["swagger_dict"]



"""Manager for askar profile multitenancy mode."""

from typing import Iterable, Optional, cast

from ..askar.profile import AskarProfile
from ..askar.profile_anon import AskarAnonCredsProfile
from ..config.injection_context import InjectionContext
from ..config.wallet import wallet_config
from ..core.profile import Profile
from ..wallet.models.wallet_record import WalletRecord
from .base import BaseMultitenantManager


class SingleWalletAskarMultitenantManager(BaseMultitenantManager):
    """Class for handling askar profile multitenancy."""

    DEFAULT_MULTITENANT_WALLET_NAME = "multitenant_sub_wallet"

    def __init__(
        self, profile: Profile, multitenant_profile: Optional[AskarProfile] = None
    ):
        """Initialize askar profile multitenant Manager.

        Args:
            profile: The base profile for this manager
            multitenant_profile: The multitenant profile for this manager
        """
        super().__init__(profile)
        self._multitenant_profile: Optional[AskarProfile] = multitenant_profile

    @property
    def open_profiles(self) -> Iterable[Profile]:
        """Return iterator over open profiles.

        Only the core multitenant profile is considered open.
        """
        if self._multitenant_profile:
            yield self._multitenant_profile

    async def get_wallet_profile(
        self,
        base_context: InjectionContext,
        wallet_record: WalletRecord,
        extra_settings: Optional[dict] = None,
        *,
        provision=False,
    ) -> Profile:
        """Get Askar profile for a wallet record.

        An object of type AskarProfile is returned but this should not be
        confused with the underlying profile mechanism provided by Askar that
        enables multiple "profiles" to share a wallet. Usage of this mechanism
        is what causes this implementation of BaseMultitenantManager.get_wallet_profile
        to look different from others, especially since no explicit clean up is
        required for profiles that are no longer in use.

        Args:
            base_context: Base context to extend from
            wallet_record: Wallet record to get the context for
            extra_settings: Any extra context settings
            provision: Whether to provision the wallet

        Returns:
            Profile: Profile for the wallet record

        """
        extra_settings = extra_settings or {}

        if not self._multitenant_profile:
            multitenant_wallet_name = base_context.settings.get(
                "multitenant.wallet_name", self.DEFAULT_MULTITENANT_WALLET_NAME
            )
            context = base_context.copy()

            sub_wallet_settings = {
                "wallet.recreate": False,
                "wallet.seed": None,
                "wallet.key": base_context.settings.get("wallet.key", ""),
                "wallet.rekey": base_context.settings.get("wallet.rekey"),
                "wallet.id": None,
                "wallet.name": multitenant_wallet_name,
                "wallet.type": "askar",
                "mediation.open": None,
                "mediation.invite": None,
                "mediation.default_id": None,
                "mediation.clear": None,
                "auto_provision": True,
            }
            context.settings = context.settings.extend(sub_wallet_settings)

            profile, _ = await wallet_config(context, provision=False)
            self._multitenant_profile = cast(AskarProfile, profile)

        profile_context = self._multitenant_profile.context.copy()

        if provision:
            await self._multitenant_profile.store.create_profile(wallet_record.wallet_id)

        extra_settings = {
            "admin.webhook_urls": self.get_webhook_urls(base_context, wallet_record),
            "wallet.askar_profile": wallet_record.wallet_id,
        }

        profile_context.settings = profile_context.settings.extend(
            wallet_record.settings
        ).extend(extra_settings)

        assert self._multitenant_profile.opened

        # return anoncreds profile if explicitly set as wallet type
        if profile_context.settings.get("wallet.type") == "askar-anoncreds":
            return AskarAnonCredsProfile(
                self._multitenant_profile.opened,
                profile_context,
                profile_id=wallet_record.wallet_id,
            )

        return AskarProfile(
            self._multitenant_profile.opened,
            profile_context,
            profile_id=wallet_record.wallet_id,
        )

    async def remove_wallet_profile(self, profile: Profile):
        """Remove the wallet profile instance.

        Args:
            profile: The wallet profile instance

        """
        await profile.remove()
